from landlab.components.craters.dig_craters import impactor
from landlab import ModelParameterDictionary

from landlab import RasterModelGrid
import numpy as np
import time
import pylab

#get the needed properties to build the grid:
input_file = './craters_params_degrade.txt'
inputs = ModelParameterDictionary(input_file)
nrows = inputs.read_int('nrows')
ncols = inputs.read_int('ncols')
dx = inputs.read_float('dx')
leftmost_elev = inputs.read_float('leftmost_elevation')
initial_slope = inputs.read_float('initial_slope')
nt = inputs.read_int('number_of_craters_per_loop')
loops = inputs.read_int('number_of_loops')

mg = RasterModelGrid(nrows, ncols, dx)
mg.set_looped_boundaries(True, True)

#create the fields in the grid
mg.create_node_array_zeros('planet_surface__elevation')
mg['node'][ 'planet_surface__elevation'] = np.load('init_topo.npy')

# Display a message
print( 'Running ...' )
start_time = time.time()

#instantiate the component:
craters_component = impactor(mg, input_file)

#perform the loops:
x = np.empty(nt)
y = np.empty(nt)
r = np.empty(nt)
slope = np.empty(nt)
angle = np.empty(nt)
az = np.empty(nt)
mass_balance = np.empty(nt)
for i in xrange(loops):
    for j in xrange(nt):
        if j == 1:
            craters_component._ycoord = 0.45*(mg.get_grid_xdimension()-mg.dx)
            craters_component._radius /= 10.
        mg = craters_component.excavate_a_crater_furbish(mg)
        x[j] = craters_component.impact_xy_location[0]
        y[j] = craters_component.impact_xy_location[1]
        r[j] = craters_component.crater_radius
        slope[j] = craters_component.surface_slope_beneath_crater
        angle[j] = craters_component.impact_angle_to_normal
        az[j] = craters_component.impactor_travel_azimuth
        mass_balance[j] = craters_component.mass_balance
        print 'Completed loop ', j
    mystring = 'craterssave'+str((i+1)*nt)
    np.save(mystring,mg['node']['planet_surface__elevation'])
    #Save the properties
    np.save(('x_'+str((i+1)*nt)),x)
    np.save(('y_'+str((i+1)*nt)),y)
    np.save(('r_'+str((i+1)*nt)),r)
    np.save(('slope_'+str((i+1)*nt)),slope)
    np.save(('angle_'+str((i+1)*nt)),angle)
    np.save(('az_'+str((i+1)*nt)),az)
    np.save(('mass_balance_'+str((i+1)*nt)),mass_balance)

#Finalize and plot
elev = mg['node']['planet_surface__elevation']
elev_r = mg.node_vector_to_raster(elev)
# Clear previous plots
pylab.figure(1)
pylab.close()
pylab.figure(1)
im = pylab.imshow(elev_r, cmap=pylab.cm.RdBu)  # display a colored image
pylab.colorbar(im)
pylab.title('Topography')

print('Done.')
print('Total run time = '+str(time.time()-start_time)+' seconds.')

pylab.show()

im=pylab.plot(elev_r[:,ncols//2])
pylab.show()


import numpy as np
from numpy.testing import assert_array_equal
from nose import with_setup
from nose.tools import assert_is

from landlab import RasterModelGrid


_GRID = None
_VALUES_AT_NODES = None


def setup_unit_grid():
    global _GRID, _VALUES_AT_NODES
    _GRID = RasterModelGrid(4, 5)
    _VALUES_AT_NODES = np.arange(20)


def setup_non_unit_grid():
    global _GRID, _VALUES_AT_NODES
    _GRID = RasterModelGrid(4, 5, 2.)
    _VALUES_AT_NODES = np.arange(20)


@with_setup(setup_unit_grid)
def test_unit_spacing():
    grads = _GRID.calculate_gradients_at_links(_VALUES_AT_NODES)
    assert_array_equal(
        grads,
        np.array([5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
                  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                 dtype=float))
    diffs = _GRID.calculate_diff_at_links(_VALUES_AT_NODES)
    assert_array_equal(grads, diffs)
    
    
@with_setup(setup_non_unit_grid)
def test_non_unit_spacing():
    grads = _GRID.calculate_gradients_at_links(_VALUES_AT_NODES)
    assert_array_equal(
        grads, np.array([5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
                         1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                        dtype=float) / _GRID.node_spacing)
    diffs = _GRID.calculate_diff_at_links(_VALUES_AT_NODES)
    assert_array_equal(grads, diffs / _GRID.node_spacing)


@with_setup(setup_unit_grid)
def test_out_array():
    grads = np.empty(31)
    rtn_grads = _GRID.calculate_gradients_at_links(_VALUES_AT_NODES, out=grads)
    assert_array_equal(
        grads,
        np.array([5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
                  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                 dtype=float))
    assert_is(rtn_grads, grads)


@with_setup(setup_unit_grid)
def test_diff_out_array():
    diff = np.empty(31)
    rtn_diff = _GRID.calculate_diff_at_links(_VALUES_AT_NODES, out=diff)
    assert_array_equal(
        diff,
        np.array([5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
                  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                 dtype=float))
    assert_is(rtn_diff, diff)



#! /usr/bin/env python

import numpy as np


def map_values_from_link_tail_node_to_link(mg, var_name):
    values_at_nodes = mg.at_node[var_name]
    mg.add_empty('link', var_name)
    values_at_links = mg.at_link[var_name]
    values_at_links[:] = values_at_nodes[mg.node_index_at_link_tail]


def map_values_from_link_head_node_to_link(mg, var_name):
    values_at_nodes = mg.at_node[var_name]
    mg.add_empty('link', var_name)
    values_at_links = mg.at_link[var_name]
    values_at_links[:] = values_at_nodes[mg.node_index_at_link_head]


def map_link_end_node_min_value_to_link(mg, var_name):
    values_at_nodes = mg.at_node[var_name]
    mg.add_empty('link', var_name)
    values_at_links = mg.at_link[var_name]
    np.minimum(values_at_nodes[mg.node_index_at_link_head],
               values_at_nodes[mg.node_index_at_link_tail],
               out=values_at_links)


def map_link_end_node_max_value_to_link(mg, var_name):
    values_at_nodes = mg.at_node[var_name]
    mg.add_empty('link', var_name)
    values_at_links = mg.at_link[var_name]
    np.maximum(values_at_nodes[mg.node_index_at_link_head],
               values_at_nodes[mg.node_index_at_link_tail],
               out=values_at_links)


def map_values_from_link_end_nodes_to_link(mg, var_name):
    values_at_nodes = mg.at_node[var_name]
    mg.add_empty('link', var_name)
    values_at_links = mg.at_link[var_name]
    values_at_links[:] = 0.5 * (values_at_nodes[mg.node_index_at_link_head] +
                                values_at_nodes[mg.node_index_at_link_tail])

def map_values_from_cell_node_to_cell(mg, var_name):
    values_at_nodes = mg.at_node[var_name]
    mg.add_empty('cell', var_name)
    values_at_cells = mg.at_cell[var_name]
    values_at_cells[:] = values_at_nodes[mg.node_index_at_cells]



#! /usr/bin/env python
"""
Unit tests for landlab.io.esri_ascii module.
"""

import unittest
import os
import numpy as np
from StringIO import StringIO

from landlab.io import read_esri_ascii, read_asc_header
from landlab.io import MissingRequiredKeyError, KeyTypeError, DataSizeError
from landlab import RasterModelGrid


_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


class TestReadEsriAsciiFileHugo(unittest.TestCase):
    def test_read_file_name(self):
        (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR,
                                                     'hugo_site.asc'))

        self.assertIsInstance(grid, RasterModelGrid)

        self.assertEqual(field.size, 55 * 76)
        self.assertEqual(field.shape, (55 * 76, ))

    def test_read_file_like(self):
        with open(os.path.join(_TEST_DATA_DIR, 'hugo_site.asc')) as asc_file:
            (grid, field) = read_esri_ascii(asc_file)

        self.assertIsInstance(grid, RasterModelGrid)

        self.assertEqual(field.size, 55 * 76)
        self.assertEqual(field.shape, (55 * 76, ))

    def test_reshape(self):
        with open(os.path.join(_TEST_DATA_DIR, 'hugo_site.asc')) as asc_file:
            (grid, field) = read_esri_ascii(asc_file, reshape=True)

        self.assertIsInstance(grid, RasterModelGrid)

        self.assertTrue(field.shape, (55, 76))


class TestReadEsriAsciiFile4x3(unittest.TestCase):
    def test_read_file_name(self):
        (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR,
                                                     '4_x_3.asc'))

        self.assertIsInstance(grid, RasterModelGrid)

        self.assertIsInstance(field, np.ndarray)
        self.assertEqual(field.shape, (4 * 3, ))
        self.assertListEqual(list(field.flat), [0., 1., 2.,
                                                3., 4., 5.,
                                                6., 7., 8.,
                                                9., 10., 11.,
                                               ])

    def test_read_file_like(self):
        with open(os.path.join(_TEST_DATA_DIR, '4_x_3.asc')) as asc_file:
            (grid, field) = read_esri_ascii(asc_file)

        self.assertIsInstance(grid, RasterModelGrid)

        self.assertIsInstance(field, np.ndarray)
        self.assertEqual(field.shape, (4 * 3, ))
        self.assertListEqual(list(field.flat), [0., 1., 2.,
                                                3., 4., 5.,
                                                6., 7., 8.,
                                                9., 10., 11.,
                                               ])

    def test_shape_mismatch(self):
        asc_file = StringIO(
            """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4.
5. 6. 7. 8.
9. 10. 11. 12.
            """)
        (grid, field) = read_esri_ascii(asc_file)
        self.assertEqual(field.size, 12)

        asc_file = StringIO(
            """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10. 11. 12.
            """)
        (grid, field) = read_esri_ascii(asc_file)
        self.assertEqual(field.size, 12)

    def test_size_mismatch(self):
        asc_file = StringIO(
            """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10.
            """)
        with self.assertRaises(DataSizeError):
            (grid, field) = read_esri_ascii(asc_file)


class TestReadEsriAsciiFileHeader(unittest.TestCase):
    def test_missing_required_key(self):
        asc_file = StringIO(
            """
nrows         4
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
            """)
        with self.assertRaises(MissingRequiredKeyError):
            header = read_asc_header(asc_file)

    def test_missing_mutex_key(self):
        asc_file = StringIO(
            """
ncols         3
nrows         4
yllcorner     2.
cellsize      10.
NODATA_value  -9999
            """)
        with self.assertRaises(MissingRequiredKeyError):
            header = read_asc_header(asc_file)

    def test_mutex_key(self):
        asc_file = StringIO(
            """
ncols         3
nrows         4
xllcenter     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
            """)
        header = read_asc_header(asc_file)
        self.assertEqual(header['xllcenter'], 1.)
        with self.assertRaises(KeyError):
            header['xllcorner']

        asc_file = StringIO(
            """
ncols         3
nrows         4
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
            """)
        header = read_asc_header(asc_file)
        self.assertEqual(header['xllcorner'], 1.)
        with self.assertRaises(KeyError):
            header['xllcenter']

    def test_missing_optional(self):
        asc_file = StringIO(
            """
ncols         3
nrows         4
xllcenter     1.
yllcorner     2.
cellsize      10.
            """)
        header = read_asc_header(asc_file)
        with self.assertRaises(KeyError):
            header['nodata_value']

    def test_case_insensitive(self):
        asc_file = StringIO(
            """
nCoLs         3
nrows         4
Xllcenter     1.
YLLCORNER     2.
CELLSIZE      10.
NODATA_value  -999
            """)
        header = read_asc_header(asc_file)
        for key in ['ncols', 'nrows', 'xllcenter', 'yllcorner', 'cellsize',
                    'nodata_value']:
            self.assertTrue(key in header)

    def test_wrong_type(self):
        asc_file = StringIO(
            """
nCoLs         3.5
nrows         4
Xllcenter     1.
YLLCORNER     2.
CELLSIZE      10.
NODATA_value  -999
            """)
        with self.assertRaises(KeyTypeError):
            header = read_asc_header(asc_file)


if __name__ == '__main__':
    unittest.main()



import numpy as np

class StreamPower(object):
    """
    UNDER DEVELOPMENT. DO NOT ATTEMPT TO USE YET. DEJH Sept 2013.
    """
    
    def __init__(self, grid, data, tstep):
        self.initialize(grid, data, tstep)
            
#This draws attention to a potential problem. It will be easy to have modules update z, but because noone "owns" the data, to forget to also update dz/dx...
#How about a built in grid utility that updates "derived" data (i.e., using only grid fns, e.g., slope, curvature) at the end of any given tstep loop?
#Or an explicit flagging system for all variables in the modelfield indicating if they have been updated this timestep. (Currently implemented)
#Or wipe the existance of any derived grid data at the end of a timestep entirely, so modules find they don't have it next timestep.
        
    def initialize(self, grid, data, tstep):
        self.m = 0.5
        self.n = 1.
        self.K = 1.e-3 * tstep
        self.unit_stream_power = grid.zeros(centering='node')
        #Flags for self-building of derived data:
        self.made_max_gradients = 0
        self.made_link_gradients = 0
        #This will us the MPD once finalized
        #Now perform checks for existance of needed data items:
        try:
            data.node_max_gradients
        except:
            try:
                data.node_max_gradients, dummy = grid.calculate_steepest_descent_on_nodes(data.elev, data.link_gradients)
                self.made_max_gradients = 1
            except:
                data.link_gradients = grid.calculate_gradients_at_active_links(data.elev)
                data.node_max_gradients, dummy = grid.calculate_steepest_descent_on_nodes(data.elev, data.link_gradients)
                self.made_max_gradients = 1
                self.made_link_gradients = 1
        try:
            data.flowacc.shape
        except:
            print 'Flow accumulation data is not available! Run will crash out.'
        
    def stream_power_erosion(self, grid, data):
        """
        """
        #Perform checks on self-build flags:
        if self.made_link_gradients:
            data.link_gradients = grid.calculate_gradients_at_active_links(data.elev)
        if self.made_max_gradients:
            data.node_max_gradients, dummy = grid.calculate_steepest_descent_on_nodes(data.elev, data.link_gradients)
        
        #Operate the main function:
        active_nodes = grid.get_active_cell_node_ids()
        node_max_gradients = data.node_max_gradients * np.where(data.node_max_gradients>0., 1., 0.) #Don't permit erosion on upward-pointing or flat surfaces
        unit_stream_power_active_nodes = data.flowacc[active_nodes]**self.m * node_max_gradients[active_nodes]**self.n
        self.unit_stream_power[active_nodes] = unit_stream_power_active_nodes
        
        data.elev -= (self.K * self.unit_stream_power)
        
        return data.elev



from landlab import RasterModelGrid
from landlab import ModelParameterDictionary
from landlab.utils import structured_grid as sgrid
from landlab.components.sed_trp_shallow_flow.transport_sed_in_shallow_flow import SurfaceFlowTransport

import time
import pylab
import numpy as np

#get the needed properties to build the grid:
inputs = ModelParameterDictionary('./stof_params_basin.txt')
nrows = inputs.read_int('nrows')
ncols = inputs.read_int('ncols')
dx = inputs.read_float('dx')
h_init = inputs.read_float('h_init')
z_boundary = inputs.read_float('z_boundary') #now the outlet height
drop_ht = inputs.read_float('drop_ht') #Now the inlet gradient
initial_slope = inputs.read_float('initial_slope')
time_to_run = inputs.read_int('run_time')

inlet_nodes = [0,1,2, ncols, 2*ncols]

mg = RasterModelGrid(nrows, ncols, dx)
mg.set_inactive_boundaries(False, True, False, True)
#mg.node_status[inlet_nodes] = 1
#mg.node_status[-5] = 1 #Fixed lip outlet
#print sgrid.node_tolink_index(mg.shape)[1]
#mg.reset_list_of_active_links()

#node_distances_to_inlet = mg.get_distances_of_nodes_to_point((mg.node_x[0], mg.node_y[0]))
#z0 = initial_slope*(np.amax(node_distances_to_inlet) - node_distances_to_inlet)
#z0[sgrid.boundary_nodes(mg.shape)] = 10.
#z0[inlet_nodes] = z0[inlet_nodes]+0.05
z0 = initial_slope*np.amax(mg.node_y) - initial_slope*mg.node_y
z0 = z0 + np.random.rand(nrows*ncols)/1000.
z0[-ncols:] = 1.

#create the fields in the grid
mg.create_node_array_zeros('planet_surface__elevation')
mg.create_node_array_zeros('planet_surface__water_depth')

#set the initial water depths
h = mg.create_node_array_zeros() + h_init
#h[sgrid.boundary_nodes(mg.shape)] = 0.
#h[inlet_nodes] = drop_ht
h[0:ncols] = drop_ht
mg['node'][ 'planet_surface__water_depth'] = h

# Set initial topography
x = mg.get_node_x_coords()
y = mg.get_node_y_coords()
zinit = mg.create_node_array_zeros()
zinit[:] = z0
#zinit[-5] = z_boundary
mg['node']['planet_surface__elevation'] = zinit

# Display a message
print( 'Running ...' )
start_time = time.time()

#instantiate the component:
transport_component = SurfaceFlowTransport(mg, './stof_params.txt')

#perform the loop:
elapsed_time = 0. #total time in simulation
while elapsed_time < time_to_run:
    timestep = transport_component.set_and_return_dynamic_timestep()
    mg = transport_component.transport_sed(elapsed_time)
    elapsed_time += timestep

#Finalize and plot
zm = mg.at_node['planet_surface__elevation']
h = mg.at_node['planet_surface__water_depth']
ddz=zm-zinit
print ddz[np.where(ddz!=0.)]
print np.amax(ddz)
    
# Get a 2D array version of the water depths and elevations
hr = mg.node_vector_to_raster(h)
zr = mg.node_vector_to_raster(zm)
dzr = mg.node_vector_to_raster(ddz)

# Clear previous plots
pylab.figure(1)
pylab.close()
pylab.figure(2)
pylab.close()

# Plot topography
pylab.figure(1)
pylab.subplot(131)
im = pylab.imshow(zr, cmap=pylab.cm.RdBu)  # display a colored image
pylab.colorbar(im)
pylab.title('Topography')
    
# Plot change in topo
pylab.figure(1)
pylab.subplot(132)
im = pylab.imshow(dzr, cmap=pylab.cm.RdBu)  # display a colored image
pylab.colorbar(im)
pylab.title('Topo change')
    
# Plot water depth
pylab.subplot(133)
im2 = pylab.imshow(hr, cmap=pylab.cm.RdBu)  # display a colored image
#pylab.clim(0, 0.25)
pylab.colorbar(im2)
pylab.title('Water depth')
    
# Display the plots
pylab.show()
print('Done.')
print('Total run time = '+str(time.time()-start_time)+' seconds.')



#! /usr/bin/env python

from landlab.grid import RasterModelGrid
from landlab.components.craters.component import CratersComponent


def main():
    import argparse
    #nmg comments
    #The argparse library provide an easy way to enter parameter values on 
    #the command line when running code.  
    #Default values are provided, but they can also be
    #provided at run time. 
    #
    #For Example, to specify the number of impacts when running:
    #python craters_component.py --impact-count 200
    #
    #to specify the number of impacts and grid size:
    #python craters_component.py --impact-count 200 --grid-size 500
    
    parser = argparse.ArgumentParser()
    parser.add_argument('--impact-count', type=int, default=128,
                        help='Number of impacts')
    parser.add_argument('--grid-size', type=int, default=200,
                        help='Number of grid rows and columns')
    parser.add_argument('--grid-spacing', type=float, default=.025,
                        help='Spacing between grid rows and columns')

    seed_parser = parser.add_mutually_exclusive_group()
    seed_parser.add_argument('--with-seed', type=int, default=1973,
                             help='Seed for random number generator')
    seed_parser.add_argument('--without-seed', action='store_true',
                             default=False,
                             help='Seed for random number generator')

    args = parser.parse_args()

    shape = (args.grid_size, args.grid_size)
    spacing = (args.grid_spacing, args.grid_spacing)
    impact_count = args.impact_count
    if args.without_seed:
        seed = None
    else:
        seed = args.with_seed

    grid = RasterModelGrid(shape[0], shape[1], spacing[0])

    craters_comp = CratersComponent(grid, seed=seed)
    for _ in xrange(impact_count):
        craters_comp.update()

    grid.imshow('node', 'planet_surface__elevation', grid_units=('km', 'km'),
                symmetric_cbar=True)


if __name__ == "__main__":
    main()



#! /usr/bin/env python

import numpy

from landlab.grid.base import ModelGrid, INTERIOR_NODE, BAD_INDEX_VALUE


def simple_poly_area(x, y):
    """
    Calculates and returns the area of a 2-D simple polygon.  Input vertices
    must be in sequence (clockwise or counterclockwise). *x* and *y* are
    arrays that give the x- and y-axis coordinates of the polygon's
    vertices.
        
    >>> import numpy as np
    >>> x = np.array([3., 1., 1., 3.])
    >>> y = np.array([1.5, 1.5, 0.5, 0.5])
    >>> simple_poly_area(x, y)
    2.0

    If the input coordinate arrays are 2D, calculate the area of each polygon.
    Note that when used in this mode, all polygons must have the same
    number of vertices, and polygon vertices are listed column-by-column.

    >>> x = np.array([[ 3.,  1.,  1.,  3.],
    ...               [-2., -2., -1., -1.]]).T
    >>> y = np.array([[1.5, 1.5, 0.5, 0.5],
    ...               [ 0.,  1.,  2.,  0.]]).T
    >>> simple_poly_area(x, y)
    array([ 2. ,  1.5])
    """
    # For short arrays (less than about 100 elements) it seems that the
    # Python sum is faster than the numpy sum. Likewise for the Python
    # built-in abs.
    return .5 * abs(sum(x[:-1] * y[1:] - x[1:] * y[:-1]) +
                    x[-1] * y[0] - x[0] * y[-1])


def calculate_link_lengths(pts, link_from, link_to):
    """
    Calculates and returns length of links between nodes.
    
    Inputs: pts - Nx2 numpy array containing (x,y) values
            link_from - 1D numpy array containing index numbers of nodes at 
                        starting point ("from") of links
            link_to - 1D numpy array containing index numbers of nodes at 
                      ending point ("to") of links
                      
    Returns: 1D numpy array containing horizontal length of each link
    
    Example:
        
        >>> pts = numpy.array([[0.,0.],[3.,0.],[3.,4.]]) # 3:4:5 triangle
        >>> lfrom = numpy.array([0,1,2])
        >>> lto = numpy.array([1,2,0])
        >>> calculate_link_lengths(pts, lfrom, lto)
        array([ 3.,  4.,  5.])
    """
    dx = pts[link_to, 0] - pts[link_from, 0]
    dy = pts[link_to, 1] - pts[link_from, 1]
    link_length = numpy.sqrt(dx * dx + dy * dy)
    return link_length
    
    
class VoronoiDelaunayGrid(ModelGrid):
    """
    This inherited class implements an unstructured grid in which cells are
    Voronoi polygons and nodes are connected by a Delaunay triangulation. Uses
    scipy.spatial module to build the triangulation.
    
    """
    #print 'VoronoiDelaunayGrid.__init__'
    
    def __init__(self, x=None, y=None, **kwds):
        
        if (x is not None) and (y is not None):
            self._initialize(x, y)
        super(VoronoiDelaunayGrid, self).__init__(**kwds)
        
        
    def _initialize(self, x, y):
        """
        Creates an unstructured grid around the given (x,y) points.
        """
        
        assert type(x)==numpy.ndarray, 'x must be a numpy array'
        assert type(y)==numpy.ndarray, 'y must be a numpy array'
        assert len(x)==len(y), 'x and y arrays must have the same size'
        
        # Make a copy of the points in a 2D array (useful for calls to geometry
        # routines, but takes extra memory space).
        pts = numpy.zeros((len(x), 2))
        pts[:,0] = x
        pts[:,1] = y
        
        # NODES AND CELLS: Set up information pertaining to nodes and cells:
        #   - number of nodes
        #   - node x, y coordinates
        #   - default boundary status 
        #   - interior and boundary nodes
        #   - nodes associated with each cell and active cell
        #   - cells and active cells associated with each node 
        #     (or BAD_VALUE_INDEX if none)
        #
        # Assumptions we make here:
        #   - all interior (non-perimeter) nodes have cells (this should be 
        #       guaranteed in a Delaunay triangulation, but there may be 
        #       special cases)
        #   - all cells are active (later we'll build a mechanism for the user
        #       specify a subset of cells as active)
        #
        self._num_nodes = len(x)
        #print x, y
        self._node_x = x
        self._node_y = y
        [self.node_status, self.interior_nodes, self.boundary_nodes] = \
                self.find_perimeter_nodes(pts)
        self._num_active_nodes = len(self.interior_nodes)
        self._num_cells = len(self.interior_nodes)
        self._num_active_cells = self.number_of_cells
        [self.node_cell, self.cell_node] = self.setup_node_cell_connectivity(
            self.node_status, self.number_of_cells)
        self.node_activecell = self.node_cell
        self.activecell_node = self.cell_node

        # ACTIVE CELLS: Construct Voronoi diagram and calculate surface area of
        # each active cell.
        from scipy.spatial import Voronoi
        vor = Voronoi(pts)
        self.active_cell_areas = numpy.zeros(self.number_of_active_cells)
        for node in self.activecell_node:
            xv = vor.vertices[vor.regions[vor.point_region[node]],0]
            yv = vor.vertices[vor.regions[vor.point_region[node]],1]
            self.active_cell_areas[self.node_activecell[node]] = simple_poly_area(xv, yv)
        
        # LINKS: Construct Delaunay triangulation and construct lists of link
        # "from" and "to" nodes.
        [self.link_fromnode, self.link_tonode, self.active_links, self.face_width] \
                = self.create_links_and_faces_from_voronoi_diagram(vor)
        self._num_links = len(self.link_fromnode)
        self._num_faces = self._num_links # temporary: to be done right!
                    
        # LINKS: Calculate link lengths
        self._link_length = calculate_link_lengths(pts, self.link_fromnode,
                                                  self.link_tonode)
                                                       
        # LINKS: inlink and outlink matrices
        self._setup_inlink_and_outlink_matrices()
        
        # ACTIVE LINKS: Create list of active links, as well as "from" and "to"
        # nodes of active links.
        self._reset_list_of_active_links()

        # LINKS: ID of corresponding face, if any
        self.link_face = (numpy.zeros(self.number_of_links, dtype=int) +
                          BAD_INDEX_VALUE)  # make the list
        face_id = 0
        for link in self.active_links:
            self.link_face[link] = face_id
            face_id += 1
            

    def find_perimeter_nodes(self, pts):
    
        # Calculate the convex hull for the set of points
        from scipy.spatial import ConvexHull
        hull = ConvexHull(pts, qhull_options='Qc') # see below why we use 'Qt'
        
        # The ConvexHull object lists the edges that form the hull. We need to
        # get from this list of edges the unique set of nodes. To do this, we
        # first flatten the list of vertices that make up all the hull edges 
        # ("simplices"), so it becomes a 1D array. With that, we can use the set()
        # function to turn the array into a set, which removes duplicate vertices.
        # Then we turn it back into an array, which now contains the set of IDs for
        # the nodes that make up the convex hull.
        #   The next thing to worry about is the fact that the mesh perimeter 
        # might contain nodes that are co-planar (that is, co-linear in our 2D 
        # world). For example, if you make a set of staggered points for a
        # hexagonal lattice using make_hex_points(), there will be some 
        # co-linear points along the perimeter. The ones of these that don't 
        # form convex corners won't be included in convex_hull_nodes, but they
        # are nonetheless part of the perimeter and need to be included in
        # the list of boundary_nodes. To deal with this, we pass the 'Qt'
        # option to ConvexHull, which makes it generate a list of coplanar
        # points. We include these in our set of boundary nodes.
        convex_hull_nodes = numpy.array(list(set(hull.simplices.flatten())))
        coplanar_nodes = hull.coplanar[:,0]
        boundary_nodes = numpy.concatenate((convex_hull_nodes, coplanar_nodes))
    
        # Now we'll create the "node_status" array, which contains the code
        # indicating whether the node is interior and active (=0) or a
        # boundary (=1). This means that all perimeter (convex hull) nodes are
        # initially flagged as boundary code 1. An application might wish to change
        # this so that, for example, some boundaries are inactive.
        node_status = numpy.zeros(len(pts[:,0]), dtype=numpy.int8)
        node_status[boundary_nodes] = 1
        
        # It's also useful to have a list of interior nodes
        interior_nodes = numpy.where(node_status==0)[0]
    
        # Return the results
        return node_status, interior_nodes, boundary_nodes
        
        
    def setup_node_cell_connectivity(self, node_status, ncells):
        """
        Creates and returns the following arrays:
            1) for each node, the ID of the corresponding cell, or
                BAD_INDEX_VALUE if the node has no cell.
            2) for each cell, the ID of the corresponding node.
            
        Inputs:
            node_status: 1D numpy array containing the boundary status code
                         for each node
            ncells: the number of cells (must equal the number of occurrences of
                    INTERIOR_NODE in node_status)
                    
        Example:
            
            >>> import landlab as ll
            >>> vdmg = VoronoiDelaunayGrid()
            >>> ns = numpy.array([1,0,0,1,0])  # 3 interior, 2 boundary nodes
            >>> [node_cell,cell_node] = vdmg.setup_node_cell_connectivity(ns, 3)
            >>> node_cell[1:3]
            array([0, 1])
            >>> node_cell[0]==BAD_INDEX_VALUE
            True
            >>> cell_node
            array([1, 2, 4])
        """
        assert ncells==numpy.count_nonzero(node_status==INTERIOR_NODE), \
               'ncells must equal number of INTERIOR_NODE values in node_status'

        cell = 0
        node_cell = numpy.ones(len(node_status), dtype=int)*BAD_INDEX_VALUE
        cell_node = numpy.zeros(ncells, dtype=int)
        for node in range(len(node_cell)):
            if node_status[node] == INTERIOR_NODE:
                node_cell[node] = cell
                cell_node[cell] = node
                cell += 1
                
        return node_cell, cell_node
        

    def create_links_from_triangulation(self, tri):
        """
        From a Delaunay Triangulation of a set of points, contained in a
        scipy.spatial.Delaunay object "tri", creates and returns:
            1) a numpy array containing the ID of the "from" node for each link
            2) a numpy array containing the ID of the "to" node for each link
            3) the number of links in the triangulation
        
        Example:
            
            >>> vdmg = VoronoiDelaunayGrid()
            >>> pts = numpy.array([[ 0., 0.],[  1., 0.],[  1., 0.87],[-0.5, 0.87],[ 0.5, 0.87],[  0., 1.73],[  1., 1.73]])
            >>> from scipy.spatial import Delaunay
            >>> dt = Delaunay(pts)
            >>> [myfrom,myto,nl] = vdmg.create_links_from_triangulation(dt)
            >>> print myfrom, myto, nl
            [5 3 4 6 4 3 0 4 1 1 2 6] [3 4 5 5 6 0 4 1 0 2 4 2] 12
        
        """
    
        # Calculate how many links there will be and create the arrays.
        #
        # The number of links equals 3 times the number of triangles minus
        # half the number of shared links. Finding out the number of shared links
        # is easy: for every shared link, there is an entry in the tri.neighbors
        # array that is > -1 (indicating that the triangle has a neighbor opposite
        # a given vertex; in other words, two triangles are sharing an edge).
        #
        num_shared_links = numpy.count_nonzero(tri.neighbors>-1)
        num_links = 3*tri.nsimplex - num_shared_links/2
        link_fromnode = numpy.zeros(num_links, dtype=int)
        link_tonode = numpy.zeros(num_links, dtype=int)
        
        # Sweep through the list of triangles, assigning "from" and "to" nodes to
        # the list of links.
        #
        # The basic algorithm works as follows. For each triangle, we will add its
        # 3 edges as links. However, we have to make sure that each shared edge
        # is added only once. To do this, we keep track of whether or not each
        # triangle has been processed yet using a boolean array called "tridone".
        # When we look at a given triangle, we check each vertex in turn. If there
        # is no neighboring triangle opposite that vertex, then we need to add the
        # corresponding edge. If there is a neighboring triangle but we haven't
        # processed it yet, we also need to add the edge. If neither condition is
        # true, then this edge has already been added, so we skip it.
        link_id = 0
        tridone = numpy.zeros(tri.nsimplex, dtype=bool)    
        for t in range(tri.nsimplex):  # loop over triangles
            for i in range(0, 3):       # loop over vertices & neighbors
                if tri.neighbors[t,i] == -1 or not tridone[tri.neighbors[t,i]]:
                    link_fromnode[link_id] = tri.simplices[t,numpy.mod(i+1,3)]
                    link_tonode[link_id] = tri.simplices[t,numpy.mod(i+2,3)]
                    link_id += 1
            tridone[t] = True
    
        # Return the results
        return link_fromnode, link_tonode, num_links
    

    def is_valid_voronoi_ridge(self, vor, n):
        
        SUSPICIOUSLY_BIG = 40000000.0
        return vor.ridge_vertices[n][0]!=-1 and vor.ridge_vertices[n][1]!=-1 \
                and numpy.amax(numpy.abs(vor.vertices[vor.ridge_vertices[n]]))<SUSPICIOUSLY_BIG

        
        
    def create_links_and_faces_from_voronoi_diagram(self, vor):
        """
        From a Voronoi diagram object created by scipy.spatial.Voronoi(),
        builds and returns:
            1) Arrays of link "from" and "to" nodes
            2) Array of link IDs for each active link
            3) Array containing with of each face
        
        Inputs: vor = a scipy.spatial.Voronoi() object that was initialized
                      with the grid nodes.
                      
        Returns four 1D numpy arrays:
            
            link_fromnode = "from" node for each link (len=num_links)
            link_tonode   = "to" node for each link (len=num_links)
            active_links  = link ID for each active link (len=num_active_links)
            face_width    = width of each face (len=num_active_links
        
        Example:
            
            >>> vdmg = VoronoiDelaunayGrid()
            >>> pts = numpy.array([[ 0., 0.],[  1., 0.],[  1.5, 0.87],[-0.5, 0.87],[ 0.5, 0.87],[  0., 1.73],[  1., 1.73]])
            >>> from scipy.spatial import Voronoi
            >>> vor = Voronoi(pts)
            >>> [fr,to,al,fw] = vdmg.create_links_and_faces_from_voronoi_diagram(vor)
            >>> fr
            array([0, 0, 0, 1, 1, 3, 3, 6, 6, 6, 4, 4])
            >>> to
            array([3, 1, 4, 2, 4, 4, 5, 4, 2, 5, 2, 5])
            >>> al
            array([ 2,  4,  5,  7, 10, 11])
            >>> fw
            array([ 0.57669199,  0.57669199,  0.575973  ,  0.57836419,  0.575973  ,
                    0.57836419])
        """
        # Each Voronoi "ridge" corresponds to a link. The Voronoi object has an
        # attribute ridge_points that contains the IDs of the nodes on either
        # side (including ridges that have one of their endpoints undefined).
        # So, we set the number of links equal to the number of ridges.
        num_links = len(vor.ridge_points)
        
        # Create the arrays for link from and to nodes
        link_fromnode = -numpy.ones(num_links, dtype=int)
        link_tonode = -numpy.ones(num_links, dtype=int)
        
        # Ridges along the perimeter of the grid will have one of their 
        # endpoints undefined. The endpoints of each ridge are contained in
        # vor.ridge_vertices, and an undefined vertex is flagged with -1.
        # Ridges with both vertices defined correspond to faces and active 
        # links, while ridges with an undefined vertex correspond to inactive
        # links. So, to find the number of active links, we subtract from the
        # total number of links the number of occurrences of an undefined
        # vertex.
        num_active_links = num_links \
                    - numpy.count_nonzero(numpy.array(vor.ridge_vertices)==-1)
        #print 'num_links=', num_links,'num_active_links=',num_active_links
        
        # Create arrays for active links and width of faces (which are Voronoi
        # ridges).
        active_links = -numpy.ones(num_active_links, dtype=int)
        face_width = -numpy.ones(num_active_links)
    
        # Loop through the list of ridges. For each ridge, there is a link, and
        # its "from" and "to" nodes are the associated "points". In addition, if
        # the ridge endpoints are defined, we have a face and an active link,
        # so we add them to our arrays as well.
        j = 0
        for i in range(num_links):
            link_fromnode[i] = vor.ridge_points[i,0]
            link_tonode[i] = vor.ridge_points[i,1]
            face_corner1 = vor.ridge_vertices[i][0]
            face_corner2 = vor.ridge_vertices[i][1]
            if self.is_valid_voronoi_ridge(vor, i):  # means it's a valid face
                dx = vor.vertices[face_corner2,0]-vor.vertices[face_corner1,0]
                dy = vor.vertices[face_corner2,1]-vor.vertices[face_corner1,1]
                face_width[j] = numpy.sqrt(dx*dx+dy*dy)
                if abs(face_width[j])>=40000.0:
                    print 'link',i,'from',link_fromnode[i],'to',link_tonode[i],'has face width',face_width[j]
                    print vor.ridge_vertices[i]
                    print vor.vertices[vor.ridge_vertices[i]]
                    from scipy.spatial import voronoi_plot_2d
                    voronoi_plot_2d(vor)
                assert face_width[j] < 40000., 'face width must be less than earth circumference!'
                active_links[j] = i
                j += 1
        #print 'active links:',active_links
        #print 'vor ridge points:',vor.ridge_points
        return link_fromnode, link_tonode, active_links, face_width





#! /usr/env/python
"""
Python implementation of ModelGrid, a class used to
create and manage grids for 2D numerical models.

First version GT, July 2010
Last modified August 2013
"""

import numpy
import warnings

from landlab.testing.decorators import track_this_method
from landlab.utils import count_repeated_values
from landlab.utils.decorators import make_return_array_immutable
from landlab.field import ModelDataFields
from . import grid_funcs as gfuncs


BAD_INDEX_VALUE = numpy.iinfo(numpy.int).max

# Map names grid elements to the ModelGrid attribute that contains the count
# of that element in the grid.
_ARRAY_LENGTH_ATTRIBUTES = {
    'node': 'number_of_nodes',
    'cell': 'number_of_cells',
    'link': 'number_of_links',
    'face': 'number_of_faces',
    'active_node': 'number_of_active_nodes',
    'active_cell': 'number_of_active_cells',
    'active_link': 'number_of_active_links',
    'active_face': 'number_of_active_faces',
}

# Define the boundary-type codes
INTERIOR_NODE = 0
FIXED_VALUE_BOUNDARY = 1
FIXED_GRADIENT_BOUNDARY = 2
TRACKS_CELL_BOUNDARY = 3

#: Indicates that a boundary node is *inactive*
INACTIVE_BOUNDARY = 4

BOUNDARY_STATUS_FLAGS_LIST = [
    FIXED_VALUE_BOUNDARY,
    FIXED_GRADIENT_BOUNDARY,
    TRACKS_CELL_BOUNDARY,
    INACTIVE_BOUNDARY,
]
BOUNDARY_STATUS_FLAGS = set(BOUNDARY_STATUS_FLAGS_LIST)


class Error(Exception):
    """
    Base class for exceptions from this module.
    """
    pass


def _sort_points_into_quadrants(x, y, nodes):
    """
    Divide points with locations given in the *x*, and *y* arrays into north,
    south, east, and west quadrants. Returns nodes contained in quadrants
    (west, east, north, south).

    >>> x = numpy.array([0, 1, 0, -1])
    >>> y = numpy.array([1, 0, -1, 0])
    >>> nodes = numpy.array([1, 2, 3, 4])
    >>> _sort_points_into_quadrants(x, y, nodes)
    (array([4]), array([2]), array([1]), array([3]))
    """
    above_x_axis = y > 0
    right_of_y_axis = x > 0
    closer_to_y_axis = numpy.abs(y) >= numpy.abs(x)

    north_nodes = nodes[above_x_axis & closer_to_y_axis]
    south_nodes = nodes[(~ above_x_axis) & closer_to_y_axis]
    east_nodes = nodes[right_of_y_axis & (~ closer_to_y_axis)]
    west_nodes = nodes[(~ right_of_y_axis) & (~ closer_to_y_axis)]

    return (west_nodes, east_nodes, north_nodes, south_nodes)


def default_axis_names(n_dims):
    _DEFAULT_NAMES = ('z', 'y', 'x')
    return _DEFAULT_NAMES[- n_dims:]


def default_axis_units(n_dims):
    return ('-', ) * n_dims


class ModelGrid(ModelDataFields):
    """
    Base class for creating and manipulating 2D structured or
    unstructured grids for numerical models.
    
    The idea is to have at least two inherited
    classes, RasterModelGrid and DelaunayModelGrid, that can create and
    manage grids. To this might be added a GenericModelGrid, which would
    be an unstructured polygonal grid that doesn't necessarily obey or
    understand the Delaunay triangulation, but rather simply accepts
    an input grid from the user. Also a HexModelGrid for hexagonal.
    """

    # Debugging flags (if True, activates some output statements)
    DEBUG_VERBOSE = False
    DEBUG_TRACK_METHODS = False

    def __init__(self, **kwds):
        #print 'ModelGrid.__init__'
        super(ModelGrid, self).__init__()
        for element_name in _ARRAY_LENGTH_ATTRIBUTES:
            array_length = self.number_of_elements(element_name)
            try:
                self.new_field_location(element_name, array_length)
            except AttributeError:
                pass

        self.axis_name = kwds.get('axis_name', default_axis_names(self.ndim))
        self.axis_units = kwds.get('axis_units', default_axis_units(self.ndim))

    def _initialize( self ):
        pass

    @property
    def ndim(self):
        """Number of spatial dimensions of the grid"""
        return 2

    @property
    def node_index_at_cells(self):
        """Node ID associated with grid cells"""
        return self.cell_node

    @property
    def active_nodes(self):
        """Node IDs of all active nodes"""
        (active_node_ids, ) = numpy.where(self.node_status != INACTIVE_BOUNDARY)
        return active_node_ids

    @property
    def node_index_at_active_cells(self):
        """Node ID associated with active grid cells"""
        (active_cell_ids, ) = numpy.where(self.node_status == INTERIOR_NODE)
        return active_cell_ids

    @property
    def active_cell_index_at_nodes(self):
        """Active cell ID associated with grid nodes"""
        return self.node_activecell

    @property
    def active_cell_index(self):
        """IDs of active cells"""
        return self.active_cells

    @property
    def node_index_at_link_head(self):
        """Node ID that defines the start of a link"""
        return self.link_fromnode

    @property
    def node_index_at_link_tail(self):
        """Node ID that defines the end of a link"""
        return self.link_tonode

    @property
    def face_index_at_links(self):
        """ID of the face associated with a link between two grid nodes"""
        return self.link_face
        
    @property
    def number_of_nodes(self):
        """Total number of nodes in the grid"""
        return self._num_nodes
    
    @property
    def number_of_cells(self):
        """Total number of cells in the grid"""
        return self._num_cells
    
    @property
    def number_of_links(self):
        """Total number of links in the grid"""
        return self._num_links
    
    @property
    def number_of_faces(self):
        """Total number of faces in the grid"""
        return self._num_faces
    
    @property
    def number_of_active_nodes(self):
        """Number of active nodes in the grid"""
        return self._num_active_nodes

    @property
    def number_of_active_cells(self):
        """Number of active cells in the grid"""
        return self._num_active_cells

    @property
    def number_of_active_links(self):
        """Number of active links in the grid"""
        return self._num_active_links

    @property
    def number_of_active_faces(self):
        """Number of active faces in the grid"""
        return self._num_active_faces

    def number_of_elements(self, element_name):
        """Return the number of elements, given by the *element_name* string
        in a grid. *element_name* must be one of:
            * node
            * cell
            * link
            * face
            * active_node
            * active_cell
            * active_link
            * active_face
        """
        try:
            return getattr(self, _ARRAY_LENGTH_ATTRIBUTES[element_name])
        except KeyError:
            raise TypeError('element name not understood')

    def get_interior_nodes(self):
        """Return node IDs of all of a grids interior nodes. Interior nodes
        are active nodes that are not on a boundary.
        """
        return numpy.where(self.node_status == INTERIOR_NODE)[0]

    @make_return_array_immutable
    def get_node_status(self):
        """
        Returns an array of node boundary-status codes.
        """
        return self.node_status

    @property
    @make_return_array_immutable
    def node_x(self):
        """X-coordinates of all nodes."""
        return self._node_x
    
    @property
    @make_return_array_immutable
    def node_y(self):
        """Y-coordinates of all nodes."""
        return self._node_y

    @make_return_array_immutable
    def node_axis_coordinates(self, axis=0):
        """
        Return node coordinates from a given *axis* (defaulting to 0). Axis
        numbering is the same as that for numpy arrays. That is, the zeroth
        axis is along the rows, and the first along the columns.
        """
        AXES = ('node_y', 'node_x')
        try:
            return getattr(self, AXES[axis])
        except IndexError:
            raise ValueError("'axis' entry is out of bounds")

    @property
    def axis_units(self):
        """A tuple of the units (as a string) for each of a grid's
        coordinates."""
        return self._axis_units

    @axis_units.setter
    def axis_units(self, new_units):
        """Set the units for each a grid's coordinates"""
        if len(new_units) != self.ndim:
            raise ValueError('length of units does not match grid dimension')
        self._axis_units = tuple(new_units)

    @property
    def axis_name(self):
        """A tuple of coordinate names for the grid"""
        return self._axis_name

    @axis_name.setter
    def axis_name(self, new_names):
        """Set the names of a grid's coordinates"""
        if len(new_names) != self.ndim:
            raise ValueError('length of names does not match grid dimension')
        self._axis_name = tuple(new_names)

    def create_node_array_zeros( self, name=None ):
        """
        Returns a 1D numpy array the same length as the number of nodes. If
        user gives optional argument 'name', we add this data to the grid with
        the specified name and return a reference to it; otherwise, we just
        create and return a 1D numpy array.
        """
        if name is None:
            return numpy.zeros(self.number_of_nodes)
        else: 
            self.add_zeros('node', name)
            return self.at_node[name]
        
    def create_active_link_array_zeros( self, name=None ):
        """
        Returns a 1D numpy array the same length as the number of nodes. If
        user gives optional argument 'name', we add this data to the grid with
        the specified name and return a reference to it; otherwise, we just
        create and return a 1D numpy array.
        """
        if name is None:
            return numpy.zeros(self.number_of_active_links)
        else: 
            self.add_zeros('link', name)
            return self.at_link[name]

    def zeros(self, **kwds):
        """
        Returns a numpy array of zeros that is the same length as the number
        of nodes in the grid. Use the *centering* keyword to return an
        array for other elements of the grid. *centering* is a string that is
        one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.
        """
        centering = kwds.pop('centering', 'node')
        try:
            return numpy.zeros(self.number_of_elements(centering), **kwds)
        except KeyError:
            raise TypeError(centering)

    def empty(self, **kwds):
        """
        Returns a numpy array of uninitialized values that is the same length
        as the number of nodes in the grid. Use the *centering* keyword to
        return an array for other elements of the grid. *centering* is a
        string that is one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.
        """
        centering = kwds.pop('centering', 'node')
        try:
            return numpy.empty(self.number_of_elements(centering), **kwds)
        except KeyError:
            raise TypeError(centering)

    def ones(self, **kwds):
        """
        Returns a numpy array of ones that is the same length as the number
        of nodes in the grid. Use the *centering* keyword to return an
        array for other elements of the grid. *centering* is a string that is
        one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.
        """
        centering = kwds.pop('centering', 'node')
        try:
            return numpy.ones(self.number_of_elements(centering), **kwds)
        except KeyError:
            raise TypeError(centering)

    def set_fixed_value_boundaries(self, node_ids):
        """
        Assignes FIXED_VALUE_BOUNDARY status to specified nodes.
        """
        self.node_status[node_ids] = FIXED_VALUE_BOUNDARY
        self._reset_list_of_active_links()

    @track_this_method
    def calculate_diff_at_links(self, node_values, out=None):
        """
        Calculates the gradient in quantity *node_values* at each active link
        in the grid.
        """
        return gfuncs.calculate_diff_at_links(self, node_values, out=out)
        
    @track_this_method
    def calculate_diff_at_active_links(self, node_values, out=None):
        """
        Calculates the differenct in quantity *node_values* at each active link
        in the grid.
        """
        return gfuncs.calculate_diff_at_active_links(self, node_values,
                                                     out=out)
        
    @track_this_method
    def calculate_gradients_at_links(self, node_values, out=None):
        """
        Calculates the gradient in quantity *node_values* at each active link
        in the grid.
        """
        return gfuncs.calculate_gradients_at_links(self, node_values, out=out)
        
    @track_this_method
    def calculate_gradients_at_active_links(self, node_values, out=None):
        """
        Calculates the gradient in quantity *node_values* at each active link
        in the grid.
        """
        return gfuncs.calculate_gradients_at_active_links(self, node_values,
                                                          out=out)
        
    @track_this_method
    def calculate_gradients_at_active_links_slow(self, s, gradient=None):
        """
        Calculates the gradient in quantity s at each active link in the grid.
        """
        if gradient==None:
            gradient = numpy.zeros(self.number_of_active_links)
            
        assert (len(gradient) == self.number_of_active_links), \
                "len(gradient)!=number_of_active_links"
                
        active_link_id = 0
        for link_id in self.active_links:
            gradient[active_link_id] = (s[self.link_tonode[link_id]]
                                        -s[self.link_fromnode[link_id]]) / \
                                        self.link_length[link_id]
            active_link_id += 1
        
        return gradient
        
    def resolve_values_on_links(self, link_values, out=None):
        """
        Resolves values provided defined on links into the x and y directions.
        Returns values_along_x, values_along_y
        """
        return gfuncs.resolve_values_on_links(self, link_values, out=out)

    def resolve_values_on_active_links(self, link_values, out=None):
        """
        Resolves values provided defined on active links into the x and y 
        directions.
        Returns values_along_x, values_along_y
        """
        return gfuncs.resolve_values_on_active_links(self, link_values, out=out)
        
    def calculate_flux_divergence_at_active_cells(self, active_link_flux, 
                                                  net_unit_flux=None):
        """
        Given an array of fluxes along links, computes the net total flux
        within each cell, divides by cell area, and stores the result in
        net_unit_flux. Overrides method of the same name in ModelGrid (nearly
        identical, but uses scalars dx and cellarea instead of variable
        link length and active cell area, respectively).
        
        The function works by calling calculate_flux_divergence_at_nodes, then
        slicing out only the values at active cells. Therefore, it is slower
        than calculate_flux_divergence_at_nodes, even though it returns a
        shorter list of numbers.
        
        The input active_link_flux should be flux of
        something (e.g., mass, momentum, energy) per unit face width, positive
        if flowing in the same direction as its link, and negative otherwise.
        There should be one value per active link. Returns an array of net
        total flux per unit area, one value per active cell (creates this
        array if it is not given as an argument).
          By convention, divergence is positive for net outflow, and negative 
        for net outflow. That's why we *add* outgoing flux and *subtract* 
        incoming flux. This makes net_unit_flux have the same sign and 
        dimensions as a typical divergence term in a conservation equation.

        In general, for a polygonal cell with $N$ sides of lengths
        Li and with surface area A, the net influx divided by cell
        area would be:
            .. math::
                {Q_{net} \over A} = {1 \over A} \sum{q_i L_i}

        For a square cell, which is what we have in RasterModelGrid,
        the sum is over 4 sides of length dx, and
        :math:`A = dx^2`, so:
            .. math::
                {Q_{net} \over A} = {1 \over dx} \sum{q_i}

        .. note::
            The net flux is defined as positive outward, negative
            inward. In a diffusion problem, for example, one would use:
                .. math::
                    {du \over dt} = \\text{source} - \\text{fd}
            where fd is "flux divergence".
            
        Example:
            
            >>> from landlab import RasterModelGrid
            >>> rmg = RasterModelGrid(4, 5, 1.0)
            >>> u = [0., 1., 2., 3., 0.,
            ...      1., 2., 3., 2., 3.,
            ...      0., 1., 2., 1., 2.,
            ...      0., 0., 2., 2., 0.]
            >>> u = numpy.array(u)
            >>> grad = rmg.calculate_gradients_at_active_links(u)
            >>> grad
            array([ 1.,  1., -1., -1., -1., -1., -1.,  0.,  1.,  1.,  1., -1.,  1.,
                    1.,  1., -1.,  1.])
            >>> flux = -grad    # downhill flux proportional to gradient
            >>> divflux = rmg.calculate_flux_divergence_at_active_cells(flux)
            >>> divflux
            array([ 2.,  4., -2.,  0.,  1., -4.])
            
        If calculate_gradients_at_active_links is called inside a loop, you can
        improve speed slightly by creating an array outside the loop. For 
        example, do this once, before the loop:
            
            >>> divflux = rmg.zeros(centering='active_cell') # outside loop
            
        Then do this inside the loop:
            
            >>> divflux = rmg.calculate_flux_divergence_at_active_cells(flux, divflux)
            
        In this case, the function will not have to create the divflux array.
            
        """
        
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid.calculate_flux_divergence_at_active_cells'
            
        assert (len(active_link_flux) == self.number_of_active_links), \
               "incorrect length of active_link_flux array"
            
        # If needed, create net_unit_flux array
        if net_unit_flux is None:
            net_unit_flux = numpy.zeros(self.number_of_active_cells)
        else:
            net_unit_flux[:] = 0.
            
        assert (len(net_unit_flux)) == self.number_of_active_cells
        
        node_net_unit_flux = self.calculate_flux_divergence_at_nodes(active_link_flux)
                
        net_unit_flux = node_net_unit_flux[self.activecell_node]
                
        return net_unit_flux
        
    def calculate_flux_divergence_at_active_cells_slow(self, active_link_flux, 
                                                  net_unit_flux=False):
        """
        Original, slower version of calculate_flux_divergence_at_active_cells, 
        using a for-loop instead of simply calling the node-based version of
        the method. Kept here as illustration of what the method is intended
        to do.
        """
        
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid.calculate_flux_divergence_at_active_cells'
            
        assert (len(active_link_flux) == self.number_of_active_links), \
               "incorrect length of active_link_flux array"
            
        # If needed, create net_unit_flux array
        if net_unit_flux==False:
            net_unit_flux = numpy.zeros(self.number_of_active_cells)
        else:
            net_unit_flux[:] = 0.
            
        assert (len(net_unit_flux))==self.number_of_active_cells
        
        # For each active link, add up the flux out of the "from" cell and 
        # into the "to" cell.
        active_link_id = 0
        for link_id in self.active_links:
            from_cell = self.node_activecell[self.link_fromnode[link_id]]
            to_cell = self.node_activecell[self.link_tonode[link_id]]
            total_flux = active_link_flux[active_link_id] * \
                         self.face_width[self.link_face[link_id]]
            #print('Flux '+str(total_flux)+' from '+str(from_cell) \
            #      +' to '+str(to_cell)+' along link '+str(link_id))
            if from_cell != BAD_INDEX_VALUE:
                net_unit_flux[from_cell] += total_flux
                #print('cell '+str(from_cell)+' net='+str(net_unit_flux[from_cell]))
            if to_cell != BAD_INDEX_VALUE:
                net_unit_flux[to_cell] -= total_flux
                #print('cell '+str(to_cell)+' net='+str(net_unit_flux[to_cell]))
            active_link_id += 1
        
        # Divide by cell area
        net_unit_flux = net_unit_flux / self.active_cell_areas
        
        return net_unit_flux

    @track_this_method
    def calculate_flux_divergence_at_nodes(self, active_link_flux, out=None):
        """
        Same as calculate_flux_divergence_at_active_cells, but works with and
        returns a list of net unit fluxes that corresponds to all nodes, rather
        than just active cells. 
        
        Note that we don't compute net unit fluxes at
        boundary nodes (which don't have active cells associated with them, and 
        often don't have cells of any kind, because they are on the perimeter), 
        but simply return zeros for these entries. The advantage is that the 
        caller can work with node-based arrays instead of active-cell-based 
        arrays.
        """
        return gfuncs.calculate_flux_divergence_at_nodes(self, active_link_flux,
                                                        out=out)
        
    def calculate_flux_divergence_at_nodes_slow(self, active_link_flux, 
                                           net_unit_flux=False):
        """
        This is the original, slower, for-loop-based version. We keep it here
        just as an illustration of what the flux divergence functions are 
        meant to do.
        """
        
        assert (len(active_link_flux) == self.number_of_active_links), \
               "incorrect length of active_link_flux array"
            
        # If needed, create net_unit_flux array
        if net_unit_flux==False:
            net_unit_flux = numpy.zeros(self.number_of_nodes)
        else:
            net_unit_flux[:] = 0.
            
        assert (len(net_unit_flux)) == self.number_of_nodes
        
        # For each active link, add up the flux out of the "from" cell and 
        # into the "to" cell.
        active_link_id = 0
        for link_id in self.active_links:
            from_node = self.link_fromnode[link_id]
            from_cell = self.node_activecell[from_node]
            to_node = self.link_tonode[link_id]
            to_cell = self.node_activecell[to_node]
            total_flux = active_link_flux[active_link_id] * \
                         self.face_width[self.link_face[link_id]]
            if from_cell != BAD_INDEX_VALUE:
                net_unit_flux[from_node] += total_flux / \
                                            self.active_cell_areas[from_cell]
            if to_cell != BAD_INDEX_VALUE:
                net_unit_flux[to_node] -= total_flux / \
                                          self.active_cell_areas[to_cell]
            active_link_id += 1
        
        return net_unit_flux
        
    @property
    @make_return_array_immutable
    def cell_areas(self):
        """
        Returns an array of grid-cell areas.

        .. note::
            Sometimes it may make sense for a grid to not always calculate
            its cell areas but, instead, only calculate them once they are
            required. In such cases, the grid class must implement a
            _setup_cell_areas_array method, which will be called the first
            time cell areas are requested.
        """
        try:
            return self.active_cell_areas
        except AttributeError:
            return self._setup_cell_areas_array()

    def get_active_cell_node_ids( self ):
        """
        Returns an integer vector of the node IDs of all active cells.
        """
        return self.activecell_node
        
    def get_active_link_connecting_node_pair(self, node1, node2):
        """
        Returns the ID number of the active link that connects the given pair of
        nodes, or None if not found.
        
        Example:
            
            >>> import landlab as ll
            >>> rmg = ll.RasterModelGrid(4, 5)
            >>> rmg.get_active_link_connecting_node_pair(8, 3)
            2
        """
        active_link = None
        for alink in xrange(0, self.number_of_active_links):
            link_connects_nodes = (
                (self.activelink_fromnode[alink] == node1 and
                 self.activelink_tonode[alink] == node2) or
                (self.activelink_tonode[alink] == node1 and
                 self.activelink_fromnode[alink] == node2))
            if link_connects_nodes:
                active_link = alink
                break
        return active_link

    @property
    def active_link_length(self):
        return self.link_length[self.active_links]

    @property
    def link_length(self):
        try:
            return self._link_length
        except AttributeError:
            return self.calculate_link_length()

    def min_active_link_length(self):
        """
        Returns the horizontal length of the shortest active link in the grid.
        """
        return numpy.amin(self.link_length[self.active_links])

    def max_active_link_length(self):
        """
        Returns the horizontal length of the longest active link in the grid.
        """
        return numpy.amax(self.link_length[self.active_links])

    def calculate_link_length(self):
        if not hasattr(self, '_link_length'):
            self._link_length = self.empty(centering='link')
        dx = (self.node_x[self.node_index_at_link_head] -
              self.node_x[self.node_index_at_link_tail])
        dy = (self.node_y[self.node_index_at_link_head] -
              self.node_y[self.node_index_at_link_tail])
        numpy.sqrt(dx ** 2 + dy **2, out=self._link_length)
        return self.link_length

    def assign_upslope_vals_to_active_links( self, u, v=[0] ):
        """
        Assigns to each active link the value of u at whichever of its
        neighbors has a higher value of v. If v is omitted, uses u for
        both.
        """
        fv = numpy.zeros(self.number_of_active_links)
        if len(v) < len(u):
            for i in xrange(0, self.number_of_active_links):
                fv[i] = max(u[self.activelink_fromnode[i]], 
                            u[self.activelink_tonode[i]] )
        else:
            for i in xrange(0, self.number_of_active_links):
                if v[self.activelink_fromnode[i]] > v[self.activelink_tonode[i]]:
                    fv[i] = u[self.activelink_fromnode[i]]
                else:
                    fv[i] = u[self.activelink_tonode[i]]
        return fv
        
    def _reset_list_of_active_links(self):
        """
        Creates or resets a list of active links. We do this by sweeping
        through the given lists of from and to nodes, and checking the status
        of these as given in the node_status list. A link is active if both its
        nodes are active interior points, or if one is an active interior and
        the other is an active boundary.
        """
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid._reset_list_of_active_links'
            
        fromnode_status = self.node_status[self.link_fromnode]
        tonode_status = self.node_status[self.link_tonode]

        active_links = (((fromnode_status == INTERIOR_NODE) & ~
                         (tonode_status == INACTIVE_BOUNDARY)) |
                        ((tonode_status == INTERIOR_NODE) & ~
                         (fromnode_status == INACTIVE_BOUNDARY)))

        (self.active_links, ) = numpy.where(active_links)

        self._num_active_links = len(self.active_links)
        self._num_active_faces = self._num_active_links
        self.activelink_fromnode = self.link_fromnode[self.active_links]
        self.activelink_tonode = self.link_tonode[self.active_links]
        
        # Set up active inlink and outlink matrices
        self._setup_active_inlink_and_outlink_matrices()

    def set_nodata_nodes_to_inactive(self, node_data, nodata_value):
        """
        Sets self.node_status to INACTIVE_BOUNDARY for all nodes whose value of
        node_data is equal to the nodata_value.
        
        Example:
            
            >>> import landlab as ll
            >>> mg = ll.RasterModelGrid(3, 4, 1.0)
            >>> mg.node_status
            array([1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1], dtype=int8)
            >>> h = numpy.array([-9999,-9999,-9999,-9999,-9999,-9999,12345.,0.,-9999,0.,0.,0.])
            >>> mg.set_nodata_nodes_to_inactive(h, -9999)
            >>> mg.node_status
            array([4, 4, 4, 4, 4, 4, 0, 1, 4, 1, 1, 1], dtype=int8)
        """
        
        # Find locations where value equals the NODATA code and set these nodes
        # as inactive boundaries.
        nodata_locations = numpy.nonzero(node_data==nodata_value)
        self.node_status[nodata_locations] = INACTIVE_BOUNDARY
        
        # Recreate the list of active cell IDs
        node_ids = numpy.array(range(0, self.number_of_nodes))
        self.activecell_node = node_ids[numpy.where(self.node_status == INTERIOR_NODE)]
        
        # Recreate the list of active links
        self._reset_list_of_active_links()
        
        
    def max_of_link_end_node_values(self, node_data):
        """
        For each active link, finds and returns the maximum value of node_data
        at either of the two ends. Use this, for example, if you want to find
        the maximum value of water depth at linked pairs of nodes (by passing
        in an array of water depth values at nodes).
        
        node_data: a 1D numpy array with length = number of nodes
        returns: a 1D numpy array of maximum values, with length = number of
            active links.
        
        Example:
            
            >>> import landlab as ll
            >>> mg = ll.RasterModelGrid(3, 4, 1.0)
            >>> h = numpy.array([2.,2.,8.,0.,8.,0.,3.,0.,5.,6.,8.,3.])
            >>> mg.max_of_link_end_node_values(h)
            array([ 2.,  8.,  6.,  8.,  8.,  3.,  3.])
        """
        return numpy.maximum(node_data[self.activelink_fromnode],
                             node_data[self.activelink_tonode])
        
    def calculate_numbers_of_node_neighbors(self):
        """
        Calculates the number of neighboring nodes for each node, and returns
        the result as a 1D numpy array. Used to find the maximum number of
        neighbors, so that inlink and outlink matrices can be dimensioned
        accordingly. Assumes that self.number_of_nodes, self.link_fromnode, and
        self.link_tonode have already been set up.
        
        Algorithm works by simply looping through all links; for each, the 
        endpoints are neighbors of one another, so we increment the number of
        neighbors for both the endpoint nodes.
        """
        num_nbrs = numpy.zeros(self.number_of_nodes, dtype=int)
        for link in range(self.number_of_links):
            num_nbrs[self.link_fromnode[link]] += 1
            num_nbrs[self.link_tonode[link]] += 1
        return num_nbrs


    def _setup_inlink_and_outlink_matrices(self):
        """
        Creates data structures to record the numbers of inlinks and outlinks
        for each node. An inlink of a node is simply a link that has the node as
        its "to" node, and an outlink is a link that has the node as its "from".
        
        We store the inlinks in an NM-row by num_nodes-column matrix called
        node_inlink_matrix. NM is the maximum number of neighbors for any node.
        
        We also keep track of the total number of inlinks and outlinks at each
        node in the num_inlinks and num_outlinks arrays.
        
        The inlink and outlink matrices are useful in numerical calculations.
        Each row of each matrix contains one inlink or outlink per node. So, if
        you have a corresponding "flux" matrix, you can map incoming or
        outgoing fluxes onto the appropriate nodes. More information on this is
        in the various calculate_flux_divergence... functions.
        
        What happens if a given node does not have two inlinks or outlinks? We
        simply put the default value -1 in this case. This allows us to use a 
        cute little trick when computing inflows and outflows. We make our 
        "flux" array one element longer than the number of links, with the last
        element containing the value 0. Thus, any time we add an influx from 
        link number -1, Python takes the value of the last element in the array,
        which is zero. By doing it this way, we maintain the efficiency that 
        comes with the use of numpy. Again, more info can be found in the 
        description of the flux divergence functions.
        
        Example:
            
        """
        
        # Find the maximum number of neighbors for any node
        num_nbrs = self.calculate_numbers_of_node_neighbors()
        self.max_num_nbrs = numpy.amax(num_nbrs)

        # Create active in-link and out-link matrices.
        self.node_inlink_matrix = - numpy.ones(
            (self.max_num_nbrs, self.number_of_nodes), dtype=numpy.int)
        self.node_outlink_matrix = - numpy.ones(
            (self.max_num_nbrs, self.number_of_nodes), dtype=numpy.int)

        # Set up the inlink arrays
        tonodes = self.link_tonode
        self.node_numinlink = numpy.bincount(tonodes,
                                             minlength=self.number_of_nodes)

        counts = count_repeated_values(self.link_tonode)
        for (count, (tonodes, link_ids)) in enumerate(counts):
            self.node_inlink_matrix[count][tonodes] = link_ids

        # Set up the outlink arrays
        fromnodes = self.link_fromnode
        self.node_numoutlink = numpy.bincount(fromnodes,
                                              minlength=self.number_of_nodes)
        counts = count_repeated_values(self.link_fromnode)
        for (count, (fromnodes, link_ids)) in enumerate(counts):
            self.node_outlink_matrix[count][fromnodes] = link_ids
                
        
    def _setup_active_inlink_and_outlink_matrices(self):
        """
        Creates data structures to record the numbers of active inlinks and 
        active outlinks for each node. These data structures are equivalent to
        the "regular" inlink and outlink matrices, except that it uses the IDs
        of active links (only).
        """
        # Create active in-link and out-link matrices.
        self.node_active_inlink_matrix = - numpy.ones(
            (self.max_num_nbrs, self.number_of_nodes), dtype=numpy.int)
        self.node_active_outlink_matrix = - numpy.ones(
            (self.max_num_nbrs, self.number_of_nodes), dtype=numpy.int)

        # Set up the inlink arrays
        tonodes = self.activelink_tonode
        self.node_numactiveinlink = numpy.bincount(
            tonodes, minlength=self.number_of_nodes)

        counts = count_repeated_values(self.activelink_tonode)
        for (count, (tonodes, active_link_ids)) in enumerate(counts):
            self.node_active_inlink_matrix[count][tonodes] = active_link_ids

        # Set up the outlink arrays
        fromnodes = self.activelink_fromnode
        self.node_numactiveoutlink = numpy.bincount(
            fromnodes, minlength=self.number_of_nodes)
        counts = count_repeated_values(self.activelink_fromnode)
        for (count, (fromnodes, active_link_ids)) in enumerate(counts):
            self.node_active_outlink_matrix[count][fromnodes] = active_link_ids
    
    def display_grid(self, draw_voronoi=False):
        """
        Displays the grid (mainly for purposes of debugging/testing and
        visual examples).
        """
        import matplotlib.pyplot as plt
        
        # Plot nodes, colored by boundary vs interior
        plt.plot(self._node_x[self.interior_nodes], 
                 self._node_y[self.interior_nodes], 'go')
        plt.plot(self._node_x[self.boundary_nodes], 
                 self._node_y[self.boundary_nodes], 'ro')
                 
        # Draw links
        for i in range(self.number_of_links):
            plt.plot([self._node_x[self.link_fromnode[i]],
                     self._node_x[self.link_tonode[i]]],
                     [self._node_y[self.link_fromnode[i]],
                     self._node_y[self.link_tonode[i]]], 'k-')
                     
        # Draw active links
        for link in self.active_links:
            plt.plot([self._node_x[self.link_fromnode[link]],
                     self._node_x[self.link_tonode[link]]],
                     [self._node_y[self.link_fromnode[link]],
                     self._node_y[self.link_tonode[link]]], 'g-')
                     
        # If caller asked for a voronoi diagram, draw that too
        if draw_voronoi!=None:
            from scipy.spatial import Voronoi, voronoi_plot_2d
            pts = numpy.zeros((self.number_of_nodes, 2))
            pts[:,0] = self._node_x
            pts[:,1] = self._node_y
            vor = Voronoi(pts)
            voronoi_plot_2d(vor)
                     
        plt.show()
        
        
    def is_boundary(self, ids, boundary_flag=None):
        """
        Check if nodes at given *ids* are boundary nodes. Use the
        *boundary_flag* to specify a particular boundary type status flag.
        """
        if boundary_flag is None:
            return ~ (self.node_status[ids] == INTERIOR_NODE)
        else:
            return self.node_status[ids] == boundary_flag
    
    def _assign_boundary_nodes_to_grid_sides(self):
        """
        For each boundary node, determines whether it belongs to the left, 
        right, top or bottom of the grid, based on its distance from the grid's
        centerpoint (mean (x,y) position). Returns lists of nodes on each of 
        the four grid sides. Assumes self.node_status, self.number_of_nodes, 
        self.boundary_nodes, self._node_x, and self._node_y have been initialized.
        
        Example:

            >>> import landlab as ll
            >>> m = ll.HexModelGrid(5, 3, 1.0)
            >>> [l,r,t,b] = m._assign_boundary_nodes_to_grid_sides()
            >>> l
            array([ 7, 12,  3], dtype=int32)
            >>> r
            array([11, 15,  6], dtype=int32)
            >>> t
            array([16, 18, 17], dtype=int32)
            >>> b
            array([0, 2, 1], dtype=int32)
        """
        # Calculate x and y distance from centerpoint
        dx = self._node_x[self.boundary_nodes] - numpy.mean(self._node_x)
        dy = self._node_y[self.boundary_nodes] - numpy.mean(self._node_y)

        return _sort_points_into_quadrants(dx, dy, self.boundary_nodes)
        
    def set_inactive_boundaries(self, bottom_is_inactive, right_is_inactive, 
                                top_is_inactive, left_is_inactive):
        """
        Handles boundary conditions by setting each of the four sides of the 
        rectangular grid to either 'inactive' or 'active (fixed value)' status.
        Arguments are booleans indicating whether the bottom, right, top, and
        left are inactive (True) or not (False).
        
        For an inactive boundary:
            - the nodes are flagged INACTIVE_BOUNDARY
            - the links between them and the adjacent interior nodes are
              inactive (so they appear on link-based lists, but not
              active_link-based lists)
              
        This means that if you call the calculate_gradients_at_active_links
        method, the inactive boundaries will be ignored: there can be no
        gradients or fluxes calculated, because the links that connect to that
        edge of the grid are not included in the calculation. So, setting a
        grid edge to INACTIVE_BOUNDARY is a convenient way to impose a no-flux
        boundary condition. Note, however, that this applies to the grid as a
        whole, rather than a particular variable that you might use in your
        application. In other words, if you want a no-flux boundary in one
        variable but a different boundary condition for another, then use 
        another method.
        
        The following example sets the top and left boundaries as inactive in a
        four-row by five-column grid that initially has all boundaries active
        and all boundary nodes coded as FIXED_VALUE_BOUNDARY (=1):
        
        >>> import landlab as ll
        >>> rmg = ll.HexModelGrid(5, 3, 1.0) # rows, columns, spacing
        >>> rmg.number_of_active_links
        30
        >>> rmg.node_status
        array([1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1], dtype=int8)
        >>> rmg.set_inactive_boundaries(False, False, True, True)
        >>> rmg.number_of_active_links
        21
        >>> rmg.node_status
        array([1, 1, 1, 4, 0, 0, 1, 4, 0, 0, 0, 1, 4, 0, 0, 1, 4, 4, 4], dtype=int8)
        """
        if self.DEBUG_TRACK_METHODS:
            print 'ModelGrid.set_inactive_boundaries'
            
        [left_edge, right_edge, top_edge, bottom_edge] = \
                self._assign_boundary_nodes_to_grid_sides()
            
        if bottom_is_inactive:
            self.node_status[bottom_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[bottom_edge] = FIXED_VALUE_BOUNDARY

        if right_is_inactive:
            self.node_status[right_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[right_edge] = FIXED_VALUE_BOUNDARY
            
        if top_is_inactive:
            self.node_status[top_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[top_edge] = FIXED_VALUE_BOUNDARY

        if left_is_inactive:
            self.node_status[left_edge] = INACTIVE_BOUNDARY
        else:
            self.node_status[left_edge] = FIXED_VALUE_BOUNDARY
        
        self._reset_list_of_active_links()

    def set_inactive_nodes(self, nodes):
        self.node_status[nodes] = INACTIVE_BOUNDARY
        self._reset_list_of_active_links()

    def get_distances_of_nodes_to_point(self, tuple_xy, get_az=0, node_subset=numpy.nan):
        """
        Returns an array of distances for each node to a provided point.
        If "get_az" is set to 'angles', returns both the distance array and an
        array of azimuths from up/north. If it is set to 'displacements', it
        returns the azimuths as a 2xnnodes array of x and y displacements.
        If it is not set, returns just the distance array.
        If "node_subset" is set as an ID, or list/array/etc of IDs method
        returns just the distance (and optionally azimuth) for that node.
        Point is provided as a tuple (x,y).
        """
        assert isinstance(tuple_xy, tuple)
        assert len(tuple_xy) == 2
        
        try:
            x_displacement = self.node_x[node_subset]-tuple_xy[0]
            y_displacement = self.node_y[node_subset]-tuple_xy[1]
        except:
            x_displacement = (self.node_x-tuple_xy[0])
            y_displacement = (self.node_y-tuple_xy[1])

        dist_array = numpy.sqrt(x_displacement*x_displacement + y_displacement*y_displacement)
        
        if get_az:
            if get_az == 'displacements':
                azimuths_as_displacements = numpy.vstack(x_displacement,y_displacement)
                del(x_displacement)
                del(y_displacement)
                return dist_array, azimuths_as_displacements
            elif get_az == 'angles':
                try:
                    angle_to_xaxis = numpy.arctan(y_displacement/x_displacement)
                except: #These cases have the impact right on a gridline.
                    try:
                        azimuth_array = numpy.empty(len(x_displacement))
                    except: #this is the single node case, impact directly N or S of the node of interest
                        if y_displacement<0:
                            azimuth_array = numpy.pi
                        else:
                            azimuth_array = 0.
                    else: #general case with whole array, with the impact right one one of the gridlines
                        nonzero_nodes = numpy.nonzero(x_displacement)
                        angle_to_xaxis = numpy.arctan(y_displacement[nonzero_nodes]/x_displacement[nonzero_nodes])
                        azimuth_array[nonzero_nodes] = numpy.where(x_displacement[nonzero_nodes]<0,1.5*numpy.pi-angle_to_xaxis,0.5*numpy.pi-angle_to_xaxis)
                    zero_nodes = numpy.where(x_displacement==0.)[0]
                    azimuth_array[zero_nodes] = numpy.where(y_displacement[zero_nodes]<0,numpy.pi,0.)
                    del(zero_nodes)
                    del(x_displacement)
                    del(y_displacement)
                    del(azimuth_array)
                    del(angle_to_xaxis)
                    del(nonzero_nodes)
                else: #the normal case
                    azimuth_array = ((1.-numpy.sign(x_displacement))*0.5)*numpy.pi + (0.5*numpy.pi-angle_to_xaxis)
                    del(x_displacement)
                    del(angle_to_xaxis)
                return dist_array, azimuth_array
            else:
                print "Option set for get_az not recognised. Should be 'displacements' or 'angles'."
        else:
            return dist_array
            
    def build_all_node_distances_azimuths_maps(self):
        """
        This function creates and stores in the grid field two nnodes*nnodes 
        arrays that map the distances and azimuths of all nodes in the grid to 
        all nodes in the grid.
        This is useful if your module needs to make repeated lookups of distances
        between the same nodes, but does potentially use up a lot of memory so
        should be used with caution.
        The map is symmetrical, so it does not matter whether rows are "from" or
        "to".
        The arrays are called:
            self.all_node_distances_map
            self.all_node_azimuths_map
        """
        
        self.all_node_distances_map = numpy.empty((self.number_of_nodes,
                                                  self.number_of_nodes))
        self.all_node_azimuths_map = numpy.empty((self.number_of_nodes,
                                                 self.number_of_nodes))
        
        node_coords = numpy.empty((self.number_of_nodes, 2))
        node_coords[:,0] = self.node_x
        node_coords[:,1] = self.node_y
        
        for i in xrange(self.number_of_nodes):
            self.all_node_distances_map[i,:], self.all_node_azimuths_map[i,:] = self.get_distances_of_nodes_to_point((node_coords[i,0],node_coords[i,1]), get_az='angles')
        
        assert numpy.all(self.all_node_distances_map >= 0.)
        
        return self.all_node_distances_map, self.all_node_azimuths_map
        

if __name__ == '__main__':
    import doctest
    doctest.testmod()



import landlab.components.craters as craters
craters = reload(craters)
from landlab import RasterModelGrid
import numpy
import time
from pylab import show, imshow, colorbar, plot

class data(object):
    '''
        This is where all the whole-grid data lives, as arrays over the various elements of the grid.
        '''
    #Data goes here!!!
    def __init__(self, grid):
        self.elev = grid.zeros(centering='node') #some data
        self.flag_already_in_the_list = grid.zeros(centering='node')
        self.craters_over_max_radius_not_plotted = grid.zeros(centering='node')
        self.impact_sequence = []

def dig_one_crater_timetest(reps, nr, nc, dx, rel_x, rel_y, radius, switch):
    '''
        This script digs a single crater, either using the crawler method (switch=0), or by sweeping the whole grid (switch=1). It is designed to time the the crater digging scripts in crater.py, with an eye to determining 1. if solving over the whole grid can be faster than the conditional crawler, and 2. if there's a crossover point for small (i.e., typical!) craters.
    '''
    mg = RasterModelGrid()
    mg.initialize(nr, nc, dx)
    vectors = data(mg)
    vectors.elev[:] = 100.
    cr = craters.impactor()

    start_time = time.time()
    for i in range(reps):
        cr._radius = radius
        #print 'Radius: ', cr._radius
        cr.set_depth_from_size()
        #print 'Depth: ', cr._depth
        cr.set_crater_volume()
        cr._xcoord = rel_x*mg.get_grid_xdimension()
        cr._ycoord = rel_y*mg.get_grid_ydimension()
        vertices_array = mg.get_nodes_around_point(cr._xcoord, cr._ycoord)
        distances_to_vertices = []
        #print vertices_array
        #print len(mg.node_x)
        for x in vertices_array:
            distances_to_vertices.append(numpy.sqrt((cr._xcoord-mg.node_x[x])**2. + (cr._ycoord-mg.node_y[x])**2.))
        cr.closest_node_index = vertices_array[numpy.argmin(distances_to_vertices)]
        cr.closest_node_elev = vectors.elev[cr.closest_node_index]
        
        cr._angle_to_horizontal = numpy.pi*0.5*3./3.
        cr._azimuth_of_travel = numpy.pi*1.5
        cr.set_crater_mean_slope_v2(mg, vectors)
        print 'Azimuth of travel: ', cr._azimuth_of_travel
        print 'Angle of ground: ', cr._surface_slope
        print 'Dip direction of ground: ', cr._surface_dip_direction
        if switch:
            cr.set_elev_change_across_grid(mg, vectors)
        else:
            cr.set_elev_change_at_pts(mg, vectors)
        print 'Impact angle to ground normal: ', cr.impactor_angle_to_surface_normal
    
    end_time = time.time()
    print('Elapsed time was %g seconds' % (end_time - start_time))
    
    elev_raster = mg.node_vector_to_raster(vectors.elev, flip_vertically=True)
    vectors.viewing_raster = elev_raster
    return cr, mg, vectors


crater, mg, vectors = dig_one_crater_timetest(1,1200,1200,0.0025, 0.5, 0.5, .75, 1)
imshow(vectors.viewing_raster)
show()

start_time = time.time()
mg, vectors_eroded, profile, xsec = craters.dig_some_craters(mg, vectors, 20000)
end_time = time.time()
print('Elapsed time was %g seconds' % ((end_time - start_time)))

imshow(vectors_eroded.viewing_raster)
show()




#! /usr/env/python
"""

This tests the overland flow and shear stress generator.

"""

import landlab
from landlab.components.overland_flow.generate_overland_flow_DEM import OverlandFlow
from landlab.components.uniform_precip.generate_uniform_precip import PrecipitationDistribution
from landlab.io import read_esri_ascii
from matplotlib import pyplot as plt
import numpy as np
import os
import time





def main():
    """
    initialize DEM
    Get storm info
    Generate hydrograph
    """
    
    start_time = time.time()
    
    dem_name = 'HalfFork.asc'
    input_file = 'input_data.txt'
    IT_FILE = os.path.join(os.path.dirname(__file__), input_file)

    # Create and initialize a raster model grid by reading a DEM
    DATA_FILE = os.path.join(os.path.dirname(__file__), dem_name)
    print('Reading data from "'+str(DATA_FILE)+'"')
    (rg, z) = read_esri_ascii(DATA_FILE)
    nodata_val=-9999
    # Modify the grid DEM to set all nodata nodes to inactive boundaries
    rg.set_nodata_nodes_to_inactive(z, nodata_val) # set nodata nodes to inactive bounds
    
    print('DEM has ' +
          str(rg.number_of_node_rows) + ' rows, ' +
          str(rg.number_of_node_columns) + ' columns, and cell size ' +
          str(rg.dx))
    

    # Select point to sample at.
    
    study_row = 110
    study_column = 150
      
    study_node = rg.grid_coords_to_node_id(study_row, study_column)


    ## Set outlet point to set boundary conditions.
    the_outlet_row = 240
    the_outlet_column = 215
    the_outlet_node = rg.grid_coords_to_node_id(the_outlet_row, the_outlet_column)

    rg.set_fixed_value_boundaries(the_outlet_node)
                                                                                                       
    # Get a 2D array version of the elevations for plotting purposes
    elev_raster = rg.node_vector_to_raster(z,True)
    
    # Everything below plots the topography and sampling points
    
    levels = []
    x_up = 1990
    while x_up !=2200:
        levels.append(x_up)
        x_up+=1
    s = plt.contourf(elev_raster, levels, colors='k')#('r','g','b'))
    plt.set_cmap('bone')
    cb = plt.colorbar()

    plt.plot([150],[109],'cs', label= 'Study Node')
    plt.plot([215],[9], 'wo', label= 'Outlet')
    plt.legend(loc=3)


    of=OverlandFlow()#IT_FILE,rg,0)
    of.initialize(rg)

    ## (-,-,-,-,how long we route overland flow, intensity in m/s, duration of storm) ##

    ## Trial 1, 10 year storm ##
    of.calculate_flow_at_one_point(rg,z,study_node,90,(7.167*(10**-6)), 2916)


    plt.show()


if __name__ == "__main__":
    main()



######### soil_moisture_driver.py ###########
##
## This component can be considered as an example for implementation
## of soil_moisture.py
##
##
##
#############################################

# Soil Moisture Driver

from soil_moisture import *

""" Create an Soil Moisture Object """
SM = SoilMoisture()

""" Initialize Soil Moisture """
SM.initialize()

SM._length = SM._iterate_storm

""" Create an instance of Storm Class """
PD = PrecipitationDistribution()
""" Initialize Storm - Create first storm """
PD.initialize()

""" Initializing OPTIONAL arrays to store values of variables and plotting """
P_ = zeros( SM._length, dtype = float )
Peff_ = zeros( SM._length, dtype = float )
ETA_ = zeros( SM._length, dtype = float )
Time_ = zeros( SM._length, dtype = float )
S_ = zeros( SM._length, dtype = float )
D_ = zeros( SM._length, dtype = float )
Ro_ = zeros( SM._length, dtype = float )

#######

""" Loop through the storms """
for i in range( 0, SM._iterate_storm ):
    if i != 0:
       """ Create a new storm """ 
       PD.update()
    """ Soil moisture dynamics """
    P = PD.storm_depth
    Tb = PD.interstorm_duration
    Tr = PD.storm_duration
    Fveg = 0.5 #self._soil_Fveg

    """ Update Soil Moisture """
    SM.update( P, Tb, Tr, Fveg)
    
    P_[i] = P
    Peff_[i] = SM.get_Peff()  # Example of using a get function
    S_[i] = SM.get_S()
    ETA_[i] = SM.get_ET()     # SM._ETA  Another example of using data stored in the object
    Ro_[i] = SM.get_Ro()
    D_[i] = SM._D             #SM.get_D() also works
    Time_[i] = SM.get_time()    

#######

""" Calculate RHS of Water Balance """
Total = (S_[SM._iterate_storm-1]-SM._si)*1000*SM._soil_pc*SM._zr+sum(ETA_)+sum(D_)+sum(Ro_)
""" Calculate percentage error in LHS and RHS of water balance equation """
Percent = (sum(Peff_)-Total)/(sum(Peff_))*100

print '\nPercent Error: ', Percent, '\n'

""" Plot output """
SM.plott( P_, S_, ETA_, D_, Ro_, Time_ )



import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import (assert_is, assert_equal, assert_raises, raises,
                        assert_true, assert_false)

from landlab import RasterModelGrid
from landlab import BAD_INDEX_VALUE


class TestRasterModelGrid(object):
    def setup(self):
        """
        These tests use a grid that 4x5 nodes::

         15------16------17------18------19
          |       |       |       |       |
          |       |       |       |       |
          |       |       |       |       |
         10------11------12------13------14
          |       |       |       |       |
          |       |       |       |       |   
          |       |       |       |       |
          5-------6-------7-------8-------9
          |       |       |       |       |
          |       |       |       |       |
          |       |       |       |       |
          0-------1-------2-------3-------4
        """
        self.num_rows = 4
        self.num_cols = 5
        self.spacing = 1.
        self.rmg = RasterModelGrid(self.num_rows, self.num_cols,
                                   dx=self.spacing)

    def test_create_grid(self):
        """
        Create a raster grid.
        """
        grid = RasterModelGrid(num_rows=4, num_cols=5)

    def test_grid_dimensions(self):
        """
        Use the default spacing of 1.
        """
        assert_equal(self.rmg.get_grid_ydimension(), self.num_rows)
        assert_equal(self.rmg.get_grid_xdimension(), self.num_cols)

    def test_grid_dimensions_non_unit_spacing(self):
        """
        Check the width of the grid. The x-dimension is the columns, and the
        y-dimension is the rows.
        """
        rmg = RasterModelGrid(4, 5, dx=2.)
        assert_equal(rmg.get_grid_ydimension(), 8.)
        assert_equal(rmg.get_grid_xdimension(), 10.)

    def test_nodes_around_point(self):
        surrounding_ids = self.rmg.get_nodes_around_point(2.1, 1.1)
        assert_array_equal(surrounding_ids, np.array([7, 12, 13, 8]))

        surrounding_ids = self.rmg.get_nodes_around_point(2.1, .9)
        assert_array_equal(surrounding_ids, np.array([2, 7, 8, 3]))

    def test_neighbor_list_with_scalar_arg(self):
        assert_array_equal(self.rmg.get_neighbor_list(6),
                           np.array([7, 11, 5, 1]))
        assert_array_equal(self.rmg.get_neighbor_list(-1),
                           np.array([-1, -1, -1, -1]))

    def test_neighbor_list_with_array_arg(self):
        assert_array_equal(self.rmg.get_neighbor_list([6, -1]),
                           np.array([[7, 11, 5, 1], [-1, -1, -1, -1]]))

    def test_neighbor_list_with_no_args(self):
        assert_array_equal(
            self.rmg.get_neighbor_list(),
            np.array([[-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [ 7, 11,  5,  1], [ 8, 12,  6,  2], [ 9, 13,  7,  3], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [12, 16, 10,  6], [13, 17, 11,  7], [14, 18, 12,  8], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1]]))

    def test_neighbor_list_boundary(self):
        """
        All of the neighbor IDs for a boundary cell are -1.
        """
        import landlab.utils.structured_grid as sgrid
        for node_id in sgrid.boundary_iter(self.rmg.shape):
            assert_array_equal(self.rmg.get_neighbor_list(node_id),
                               np.array([-1, -1, -1, -1]))

    def test_node_x(self):
        assert_array_equal(self.rmg.node_x,
                           np.array([0., 1., 2., 3., 4.,
                                     0., 1., 2., 3., 4.,
                                     0., 1., 2., 3., 4.,
                                     0., 1., 2., 3., 4.]))

    def test_node_y(self):
        assert_array_equal(self.rmg.node_y,
                              np.array([0., 0., 0., 0., 0.,
                                        1., 1., 1., 1., 1.,
                                        2., 2., 2., 2., 2.,
                                        3., 3., 3., 3., 3.]))

    @raises(ValueError)
    def test_node_x_is_immutable(self):
        self.rmg.node_x[0] = 0

    @raises(ValueError)
    def test_node_y_is_immutable(self):
        self.rmg.node_y[0] = 0

    def test_node_axis_coordinates(self):
        assert_is(self.rmg.node_axis_coordinates(axis=0).base,
                  self.rmg.node_y.base)
        assert_is(self.rmg.node_axis_coordinates(axis=1).base,
                  self.rmg.node_x.base)
        assert_is(self.rmg.node_axis_coordinates(axis=-1).base,
                  self.rmg.node_x.base)
        assert_is(self.rmg.node_axis_coordinates(axis=-2).base,
                  self.rmg.node_y.base)

    def test_diagonal_list(self):
        assert_array_equal(self.rmg.get_diagonal_list(6),
                           np.array([12, 10, 0, 2]))
        assert_array_equal(self.rmg.get_diagonal_list(-1),
                           np.array([-1, -1, -1, -1]))
        assert_array_equal(self.rmg.get_diagonal_list([6, -1]),
                           np.array([[12, 10, 0, 2], [-1, -1, -1, -1]]))
        assert_array_equal(
            self.rmg.get_diagonal_list(),
            np.array([[-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1],
                        [-1, -1, -1, -1], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [12, 10,  0,  2], [13, 11,  1,  3],
                        [14, 12,  2,  4], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [17, 15,  5,  7], [18, 16,  6,  8],
                        [19, 17,  7,  9], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1],
                        [-1, -1, -1, -1], [-1, -1, -1, -1]]))

    def test_diagonal_list_boundary(self):
        assert_array_equal(self.rmg.get_diagonal_list(0),
                           np.array([-1, -1, -1, -1]))

    def test_is_interior(self):
        for cell_id in [0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 16, 17, 18, 19]:
            assert_false(self.rmg.is_interior(cell_id))
        for cell_id in [6, 7, 8, 11, 12, 13]:
            assert_true(self.rmg.is_interior(cell_id))

    def test_get_interior_cells(self):
        assert_array_equal(self.rmg.get_active_cell_node_ids(),
                           np.array([6, 7, 8, 11, 12, 13]))

    def test_active_links(self):
        assert_equal(self.rmg.number_of_active_links, 17)
        assert_array_equal(self.rmg.active_links,
                           np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                     19, 20, 21, 22, 23, 24, 25, 26]))

    def test_active_link_fromnode(self):
        assert_array_equal(self.rmg.activelink_fromnode,
                           np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                     5, 6, 7, 8, 10, 11, 12, 13]))

    def test_active_link_tonode(self):
        assert_array_equal(self.rmg.activelink_tonode,
                           np.array([6, 7, 8, 11, 12, 13, 16, 17, 18,
                                     6, 7, 8, 9, 11, 12, 13, 14]))

    def test_active_link_num_inlink(self):
        assert_array_equal(self.rmg.node_numactiveinlink,
                           np.array([0, 0, 0, 0, 0,
                                     0, 2, 2, 2, 1,
                                     0, 2, 2, 2, 1,
                                     0, 1, 1, 1, 0]))

    def test_active_link_num_outlink(self):
        assert_array_equal(self.rmg.node_numactiveoutlink,
                           np.array([0, 1, 1, 1, 0,
                                     1, 2, 2, 2, 0,
                                     1, 2, 2, 2, 0,
                                     0, 0, 0, 0, 0]))

    def test_active_inlink_matrix(self):
        assert_array_equal(
            self.rmg.node_active_inlink_matrix,
            np.array([[-1, -1, -1, -1, -1,
                       -1,  0,  1,  2, -1,
                       -1,  3,  4,  5, -1,
                       -1,  6, 7,  8, -1],
                      [-1, -1, -1, -1, -1,
                       -1,  9, 10, 11, 12,
                       -1, 13, 14, 15, 16,
                       -1, -1, -1, -1, -1]]))

    def test_active_outlink_matrix(self):
        assert_array_equal(
            self.rmg.node_active_outlink_matrix,
            np.array([[-1,  0,  1,  2, -1,
                       -1,  3,  4,  5, -1,
                       -1,  6,  7,  8, -1,
                       -1, -1, -1, -1, -1],
                      [-1, -1, -1, -1, -1,
                        9, 10, 11, 12, -1,
                       13, 14, 15, 16, -1,
                       -1, -1, -1, -1, -1]]))

    def test_active_node_links_scalar_interior(self):
        assert_array_equal(
            self.rmg.active_node_links([6]),
            np.array([[0, 9, 3, 10]]).T)

    def test_active_node_links_scalar_boundary(self):
        assert_array_equal(
            self.rmg.active_node_links([1]),
            np.array([[-1, -1, 0, -1]]).T)

    def test_active_node_with_array_arg(self):
        assert_array_equal(
            self.rmg.active_node_links([6, 7]),
            np.array([[0,  9, 3, 10],
                      [1, 10, 4, 11]]).T)

    def test_active_node_links_with_no_args(self):
        assert_array_equal(
            self.rmg.active_node_links(),
            np.array([[-1, -1, -1, -1, -1, -1,  0,  1,  2, -1,
                       -1,  3,  4,  5, -1, -1,  6,  7,  8, -1],
                      [-1, -1, -1, -1, -1, -1,  9, 10, 11, 12,
                       -1, 13, 14, 15, 16, -1, -1, -1, -1, -1],
                      [-1,  0,  1,  2, -1, -1,  3,  4,  5, -1,
                       -1,  6,  7,  8, -1, -1, -1, -1, -1, -1],
                      [-1, -1, -1, -1, -1,  9, 10, 11, 12, -1,
                       13, 14, 15, 16, -1, -1, -1, -1, -1, -1]]))

    def test_link_fromnode(self):
        assert_array_equal(
            self.rmg.link_fromnode,
            np.array(
                [0, 1, 2, 3, 4, 5, 6, 7,  8,  9, 10, 11, 12, 13, 14,
                 0, 1, 2, 3, 5, 6, 7, 8, 10, 11, 12, 13, 15, 16, 17, 18]))

    def test_link_tonode(self):
        assert_array_equal(
            self.rmg.link_tonode,
            np.array(
                [5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                 1, 2, 3, 4, 6,  7,  8,  9, 11, 12, 13, 14, 16, 17, 18, 19]))

    def test_link_num_inlink(self):
        assert_array_equal(self.rmg.node_numinlink,
                           np.array([0, 1, 1, 1, 1,
                                     1, 2, 2, 2, 2,
                                     1, 2, 2, 2, 2,
                                     1, 2, 2, 2, 2]))

    def test_link_num_outlink(self):
        assert_array_equal(self.rmg.node_numoutlink,
                           np.array([2, 2, 2, 2, 1,
                                     2, 2, 2, 2, 1,
                                     2, 2, 2, 2, 1,
                                     1, 1, 1, 1, 0]))

    def test_node_inlink_matrix(self):
        assert_array_equal(self.rmg.node_inlink_matrix,
                           np.array([[-1, -1, -1, -1, -1,
                                       0,  1,  2,  3,  4,
                                       5,  6,  7,  8,  9,
                                      10, 11, 12, 13, 14],
                                     [-1, 15, 16, 17, 18,
                                      -1, 19, 20, 21, 22,
                                      -1, 23, 24, 25, 26,
                                      -1, 27, 28, 29, 30]]))

    def test_node_outlink_matrix(self):
        assert_array_equal(self.rmg.node_outlink_matrix,
                           np.array([[ 0,  1,  2,  3,  4,
                                       5,  6,  7,  8,  9,
                                      10, 11, 12, 13, 14,
                                      -1, -1, -1, -1, -1],
                                     [15, 16, 17, 18, -1,
                                      19, 20, 21, 22, -1,
                                      23, 24, 25, 26, -1,
                                      27, 28, 29, 30, -1]]))

    def test_node_links_with_scalar_interior(self):
        assert_array_equal(self.rmg.node_links([6]),
                           np.array([[1, 19, 6, 20]]).T)

    def test_node_links_with_scalar_boundary(self):
        assert_array_equal(self.rmg.node_links([1]),
                           np.array([[-1, 15, 1, 16]]).T)

    def test_node_links_with_array_arg(self):
        assert_array_equal(self.rmg.node_links([6, 7]),
                           np.array([[1, 19, 6, 20], [2, 20, 7, 21]]).T)

    def test_node_links_with_no_args(self):
        assert_array_equal(
            self.rmg.node_links(),
            np.array([[-1, -1, -1, -1, -1,  0,  1,  2,  3,  4,
                        5,  6,  7,  8,  9, 10, 11, 12, 13, 14],
                      [-1, 15, 16, 17, 18, -1, 19, 20, 21, 22,
                       -1, 23, 24, 25, 26, -1, 27, 28, 29, 30],
                      [ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9,
                       10, 11, 12, 13, 14, -1, -1, -1, -1, -1],
                      [15, 16, 17, 18, -1, 19, 20, 21, 22, -1,
                       23, 24, 25, 26, -1, 27, 28, 29, 30, -1]]))

    def test_link_face(self):
        BAD = BAD_INDEX_VALUE
        assert_array_equal(self.rmg.link_face,
                           np.array([BAD, 0, 1, 2, BAD,
                                     BAD, 3, 4, 5, BAD,
                                     BAD, 6, 7, 8, BAD,
                                     BAD, BAD, BAD, BAD,
                                     9, 10, 11, 12,
                                     13, 14, 15, 16,
                                     BAD, BAD, BAD, BAD]))

    def test_grid_coords_to_node_id_with_scalar(self):
        assert_equal(self.rmg.grid_coords_to_node_id(3, 4), 19)

    def test_grid_coords_to_node_id_with_array(self):
        assert_array_equal(self.rmg.grid_coords_to_node_id((3, 2), (4, 1)),
                           np.array([19, 11]))

    def test_grid_coords_to_node_id_outside_of_grid(self):
        assert_raises(ValueError, self.rmg.grid_coords_to_node_id, 5, 0)

    def test_create_diagonal_list(self):
        self.rmg.create_diagonal_list()

        assert_array_equal(
            self.rmg.get_diagonal_list(),
            np.array([[-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [-1, -1, -1, -1], [12, 10,  0,  2], [13, 11,  1,  3],
                      [14, 12,  2,  4], [-1, -1, -1, -1], [-1, -1, -1, -1], [17, 15,  5,  7],
                      [18, 16,  6,  8], [19, 17,  7,  9], [-1, -1, -1, -1], [-1, -1, -1, -1],
                      [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1], [-1, -1, -1, -1]]))





import unittest

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np

import landlab


class TestImshowGrid(unittest.TestCase):
    def test_imshow_grid(self):
        rmg = landlab.RasterModelGrid(4, 5)

        pp = PdfPages('test.pdf')

        values = np.arange(rmg.number_of_nodes)
        landlab.plot.imshow_grid(rmg, values, values_at='node', limits=(0, 20))
        pp.savefig()

        plt.clf()
        values = np.arange(rmg.number_of_cells)
        landlab.plot.imshow_grid(rmg, values, values_at='cell',
                                 symmetric_cbar=True)
        pp.savefig()

        plt.clf()
        values = np.arange(rmg.number_of_cells)
        landlab.plot.imshow_grid(rmg, values, values_at='active_cell')
        pp.savefig()

        pp.close()

if __name__ == '__main__':
    unittest.main()



## Evaluating rainfall and fire time series for overlap 
##
## Runs the fire and rainfall generators for 100 years based
## on input values selected by the user. Right now, this model
## is designed to run a simulation for central Colorado, based on
## threshold information calculated by Cannon et al., 2008 for the
## Coal Seam Fire and fire recurrence from Moody and Martin, 2001 
## for the Buffalo Creek Fire.
##
## Current design not necessarily applicable to other study areas.
##
## Written by Jordan Marie Adams
## Last updated: October 29, 2013

import os
from matplotlib import pyplot as plt
from landlab.components.uniform_precip.generate_uniform_precip import PrecipitationDistribution
from landlab.components.fire_generator.generate_fire import FireGenerator
import numpy as np
from math import ceil

## INPUT TXT FILE WITH NECESSARY PARAMETERS ##
filename = os.path.join(os.path.dirname(__file__), 'fireraininput.txt')

## INITIALIZING THE CLASSES IN LANDLAB ##

Rain = PrecipitationDistribution()
Rain.initialize(filename)
Rain.get_storm_time_series() ## UNITS IN DAYS
storm= Rain.storm_time_series

Fire = FireGenerator()
Fire.initialize(filename)
Fire.get_scale_parameter() 
Fire.generate_fire_time_series()
fires = Fire.fire_events

## FUNCTIONS TO GET POTENTIAL EROSION EVENTS

## set_threshold() ##
## 
## GETS THRESHOLD BASED ON
## CANNON ET AL., 2008 RELATIONSHIP
## FOR THE COAL SEAM FIRE, EAST OF
## DENVER, COLORADO

def set_threshold(arr): 
    for event in arr:
        indx = arr.index(event)
        D = arr[indx][1] - arr[indx][0]
        I = 6.5*D**(-0.7)
        arr[indx].append(I)
    return arr
    
## get_max_intensity() ##
## 
## GETS THE MAXIMUM INTENSITY
## NEEDED FOR PLOTTING. 

def get_max_intensity(storm):
    intensity_list = []
    for intensity in storm:
        indx = storm.index(intensity)
        intensity_list.append(storm[indx][2])
    max_int = max(intensity_list)
    max_intensity = int(ceil(max_int/100.0))*100
    return max_intensity

## threshold_exceed()
##
## FINDS STORM EVENTS WHERE THE
## ACTUAL STORM DURATION AND INTENSITY
## EXCEEDS THE THRESHOLD 

def threshold_exceed(storm_arr):
    threshold_exceeded = []
    for each in storm_arr:
        index_each = storm_arr.index(each)
        if storm_arr[index_each][2] >= storm_arr[index_each][3]:
            threshold_exceeded.append(storm_arr[index_each])
    return threshold_exceeded

##  find_overlap()
## 
## THIS FUNCTION TAKES THE FIRE_ARRAY
## AND THRESHOLD ARRAY AND COMPARES THE TIMELINES
## TO FIND WHERE THERE IS A OVERLAP AND A 
## POTENTIALLY EROSION-INDUCING STORM
    
def find_overlap(fire_arr, rain_arr):
    overlap = []
    for item in fire_arr:
        i=0
        indx = fire_arr.index(item)
        start_fire = fire_arr[indx][0]
        end_fire = fire_arr[indx][1]
        while i < len(rain_arr):
            if rain_arr[i][0] < start_fire and rain_arr[i][1] > start_fire and rain_arr[i][1] < end_fire:
                overlap.append(rain_arr[i])
            if rain_arr[i][0] > start_fire and rain_arr[i][1] < end_fire:
                overlap.append(rain_arr[i])
            if rain_arr[i][0] < end_fire and rain_arr[i][1] > end_fire:
                overlap.append(rain_arr[i])
            i+=1
    return overlap
    
## write_to_file() 
##
## Takes any arrays generated prior and 
## writes values to .txt files so that
## additional plots can be made elsewhere if needed.
    
def write_to_file(fire_arr, storm_arr, exceed_arr, firetxt, stormtxt, exceedtxt):
    np.savetxt(firetxt, fire_arr)
    np.savetxt(stormtxt, storm_arr)
    np.savetxt(exceedtxt, exceed_arr)
    
## CALLING THE FUNCTIONS TO DO ALL THE WORK....
    
set_threshold(storm)
max_intense = get_max_intensity(storm)         
threshold_exceeded = threshold_exceed(storm)
thresh = find_overlap(fires, threshold_exceeded)

## create_thresh_plot()
## 
## TAKES THE DIFFERENT ARRAYS (FIRES, STORMS
## AND THRESHOLD EXCEEDENCE) AND PLOTS THEM 
## TO SHOW POTENTIALLY EROSION-INDUCING STORMS

def create_thresh_plot(fire_arr,storm_arr, exceed_arr, max_intense=1000):
    plt.figure(1)
    plt.xlabel('Time (years)', fontsize=20) ## x axis label
    plt.ylabel('Rainfall Intensity (mm/day)', fontsize=20) ## y axis label
    plt.title('Randomly Generated Rainfall Time Series', fontsize=24) ## chart title
    ax = plt.gca()
    tick_locations=[0, 3652.42, 7304.84, 10957.26, 14609.68, 18262.1, 21914.52, 25566.96, 29219.36, 32871.78, 36524.2]#labels = range(ticks.size)
    tick_labels=[0,10,20,30,40,50,60,70,80,90,100]
    plt.xticks(tick_locations, tick_labels)
    ax.tick_params(labelsize=16)
    plt.ylim(ymin=0, ymax=max_intense)
    plt.xlim(0, 36524.2)
    for f in fire_arr:
        y = fire_arr.index(f)
        start = fire_arr[y][0]
        end = fire_arr[y][0] + 1.0
        plt.broken_barh([(start, 1)], ((max_intense-200),max_intense), label='Fire', color='orange')

    for s in storm_arr:
        x = storm_arr.index(s) ##creates rainfall graph. x=length of storm. y=intensity
        start = storm_arr[x][0]
        end = storm_arr[x][1] - storm_arr[x][0]
        plt.broken_barh([(start, end)], (0,storm_arr[x][2]), label='Rain', color = 'blue') ## for each Tr period
    
    for t in exceed_arr:
        z = exceed_arr.index(t)
        start = exceed_arr[z][0]
        end = exceed_arr[z][1] - exceed_arr[z][0]
        plt.broken_barh([(start, end)], (0,exceed_arr[z][2]), label='Threshold Exceeded', color='red')
    plt.show()
        
create_thresh_plot(fires, storm, thresh, max_intense)
##write_to_file(fires, storm, thresh, fire_output.txt, storm_output.txt, thresh_output.txt)



########## rainfall_driver.py ###############
##
## This component generates rainfall  
## events based on statistical distributions.
##
## No particular units must be used, but it was
## written with the storm units in hours (hr)
## and depth units in millimeters (mm)
##
##
##
## Written by Jordan Adams, 2013.
###########################################

import os

import numpy as np 
import random
from landlab import ModelParameterDictionary

_DEFAULT_INPUT_FILE = os.path.join(os.path.dirname(__file__),
                                  'preciptest.in')

class PrecipitationDistribution:

    def __init__(self):
        '''All initial values are set to zero until initalized
        using the initialize() method which reads in data using
        the ModelParameter Dictionary and sets the random variables
        according to their respective distribution'''
        self.mean_storm = 0.0
        self.mean_intensity = 0.0
        self.mean_interstorm = 0.0
        self.mean_storm_depth = 0.0
        

        '''Storm_duration, interstorm_duration, storm_depth and intensity
        are not read in, rather, they are generated through initialize() and
        updated using the update() method'''
        self.storm_duration = 0.0
        self.interstorm_duration = 0.0
        self.storm_depth = 0.0
        self.intensity = 0.0
        
        '''TIME VARIABLES'''
        self.storm_time_series =[]


    def initialize(self, input_file=None):
        MPD = ModelParameterDictionary()
        '''
        We imported methods from ModelParameterDictionary
        to read the parameters from the input file.
 
        Necessary parameters to run this code include:
    
            mean_storm (type: float)
            mean_intensity (type: float)
            mean_interstorm (type: float)
    
        mean_storm_depth is calculated using mean storm and intensity data

        The random variables of storm_duration, interstorm_duration, storm_depth
        and intensity are found using methods declared later in the class
        '''
        if input_file is None:
            input_file = _DEFAULT_INPUT_FILE
        MPD.read_from_file(input_file)
        
        
        
        self.mean_storm = MPD.read_float( 'MEAN_STORM' )
        self.mean_intensity = MPD.read_float( 'MEAN_INTENSITY' )
        self.mean_interstorm = MPD.read_float( 'MEAN_INTERSTORM' )
        self.mean_storm_depth = self.mean_intensity * self.mean_storm

        self.storm_duration = self.get_precipitation_event_duration()
        self.interstorm_duration = self.get_interstorm_event_duration()
        self.storm_depth = self.get_storm_depth()
        self.intensity = self.get_storm_intensity()
        
        self.run_time = MPD.read_float('RUN_TIME')
        self.delta_t = MPD.read_int('DELTA_T')


    def update(self):
        
        '''This method updates the variables of storm_duration, interstorm_duration,
        storm_depth and intensity that were found in initialize() if needed.'''
        
        self.storm_duration = self.get_precipitation_event_duration()
        self.interstorm_duration = self.get_interstorm_event_duration()
        self.storm_depth = self.get_storm_depth()
        self.intensity = self.get_storm_intensity()
        
    def get_precipitation_event_duration(self):
        '''
    This method is the storm generator.
    
    This method has one argument: the mean_storm parameter.
    (In Eagleson (1978), this parameter was called Tr.)
    
    It finds a random storm_duration value
    based on the poisson distribution about the mean.
    This is accomplished using the expovariate function
    from the 'random' standard library.
    Additionally, it is rounded to contain 4 significant figures, 
    for neatness. 
    
    The if-else statement is very important here. Values of 0
    can exist in the Poission distribution, but it does not make
    sense to have 0 duration storms, so to avoid that,
    we return a storm duration IF it is greater than 0,
    otherwise, recursion is employed to re-call the storm
    generator function and get a new value.
    '''
        storm = round(random.expovariate(1/self.mean_storm),2)
        while storm == 0: 
            storm = round(random.expovariate(1/self.mean_storm),2)
        self.storm_duration = storm
        return self.storm_duration

    
    
    def get_interstorm_event_duration(self):
        ''' 
    This method is the interstorm duration generator
    
    This method takes one argument, the mean_interstorm parameter.
    (In Eagleson (1978), this parameter was called Tb.)
    
    This method is modeled identically to get_precipitation_event_duration()
    
    This method finds a random value for interstorm_duration
    based on the poisson distribution about the mean.
    This is accomplished using the expovariate function
    from the 'random' standard library.
    Additionally, it is rounded to contain 4 significant figures, for neatness. 
    
    The if-else statement is very important here. Values of 0
    can exist in the Poission distribution, but it does not make
    sense to have 0 hour interstorm durations (DOES IT NOT?)
    To avoid 0 hour interstorm durations, we return a 
    interstorm duration IF it is greater than 0,
    otherwise, recursion is employed to re-call the interstorm
    duration generator function and get a new value. '''
    
        interstorm = round(random.expovariate(1/self.mean_interstorm),2)
        while interstorm == 0:
            interstorm = round(random.expovariate(1/self.mean_interstorm),2)
        self.interstorm_duration = interstorm
        return self.interstorm_duration
     
                            
    def get_storm_depth(self):
        '''  
        This method is the storm depth generator.
    Storm depth is used to generate a realistic 
    intensity for different storm events.
    
    (In Eagleson (1978) this parameter was called 'h')
    
    This method requires storm_duration, mean_storm duration
    and the mean_storm_depth. Storm_duration is generated through
    the initialize() or update() method. mean_storm and mean_storm_depth
    are read in using the ModelParameterDictionary.
    
    Numpy has a random number generator to get values
    from a given Gamma distribution. It takes two arguments,
    alpha (or the shape parameter), which is the generated over the mean event
    and beta (or the scale parameter), which is the mean value
    These are all arguments in the function, which returns storm depth.
    '''
        
        shape_parameter = (self.storm_duration/self.mean_storm)
        scale_parameter = (self.mean_storm_depth)
        self.storm_depth = np.random.gamma(shape_parameter, scale_parameter)
        return self.storm_depth

    
    def get_storm_intensity(self):
        '''  
        This method draws storm intensity out of the storm depth
    generated by get_storm_depth. 
    
    This method requires the storm_depth and storm_duration
    and is the same as the parameter ('i') in Eagleson (1978), but instead of
    being drawn from Poission, this is drawn from the Gamma distribution
    of ('h'), as h = i*Tr. 
    
                            '''
        self.intensity = self.storm_depth / self.storm_duration
        return self.intensity


    def get_storm_time_series(self):
        '''
        This method creates a time series of storms based on storm_duration, and
        interstorm_duration. From these values it will calculate a complete
        time series.
        
        The storm_time_series returned by this method is made up of sublists, each comprising of three
        sub-parts (e.g. [[x,y,z], [a,b,c]]) where x and a are the beginning times of a precipitation 
        event, y and b are the ending times of the precipitation event and z and c represent the
        average intensity (mm/hr) of the storm lasting from x to y and a to be, respectively. 
        '''
        
        
        ''' This statement creates the first storm/interstorm pair and event.'''
        storm = self.get_precipitation_event_duration()
        self.get_storm_depth()
        intensity = self.get_storm_intensity()
        self.storm_time_series.append([0, storm, intensity])

        '''
        There is a helper and an iterator. Open to suggestions on how to make this 
        sleeker. helper keeps track of storm events so that we can properly adjust the time for the 
        storm_time_series list. 
        
        Storm iterator makes sure that the run goes through to the specified time, either read in as 'run_time' by
        the ModelParameterDictionary or specified the fire time this method is called.
        
        '''
        storm_helper = storm
        storm_iterator = storm
        while storm_iterator <= self.run_time:
            next_storm_start  = storm_helper + round(self.get_interstorm_event_duration(),2)
            next_storm_end = next_storm_start + round(self.get_precipitation_event_duration(),2)
            intensity = round(self.get_storm_intensity(),2)
            self.get_storm_depth()
            self.storm_time_series.append([next_storm_start, next_storm_end, intensity])
            storm_iterator = storm_helper
            storm_helper = next_storm_end
            storm_iterator = storm_helper
        return self.storm_time_series
            



#! /usr/bin/env python

import numpy as np

from landlab import RasterModelGrid
from landlab.components.craters.component import CratersComponent
import landlab.components.flexure as flexure


def create_planet_surface_with_bulge(grid):
    z = grid.field_values('node', 'planet_surface__elevation').view()
    z.shape = grid.shape

    (y, x) = np.meshgrid(np.linspace(0, np.pi * .5, grid.shape[0]),
                         np.linspace(0, np.pi * .5, grid.shape[1]))
    (x0, y0) = (np.pi / 3, np.pi / 8)
    np.sin((x - x0) ** 2 + (y - y0) ** 2, out=z)
    z *= .1


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--impact-count', type=int, default=128,
                        help='Number of impacts')
    parser.add_argument('--grid-size', type=int, default=200,
                        help='Number of grid rows and columns')
    parser.add_argument('--grid-spacing', type=float, default=.025,
                        help='Spacing between grid rows and columns')

    seed_parser = parser.add_mutually_exclusive_group()
    seed_parser.add_argument('--with-seed', type=int, default=1973,
                             help='Seed for random number generator')
    seed_parser.add_argument('--without-seed', action='store_true',
                             default=False,
                             help='Seed for random number generator')
    args = parser.parse_args()

    shape = (args.grid_size, args.grid_size)
    spacing = (args.grid_spacing, args.grid_spacing)
    impact_count = args.impact_count
    if args.without_seed:
        seed = None
    else:
        seed = args.with_seed

    grid = RasterModelGrid(shape[0], shape[1], spacing[0])

    craters_comp = CratersComponent(grid, seed=seed)
    flexure_comp = flexure.FlexureComponent(
        grid,
        map_vars={
            'node': {
                'lithosphere__elevation':
                    'planet_surface__elevation',
                'planet_surface_sediment__deposition_increment':
                    'planet_surface__elevation_increment'
            }
        }
    )

    create_planet_surface_with_bulge(grid)

    for _ in xrange(impact_count):
        craters_comp.update()
        flexure_comp.update()

    grid.imshow('node', 'planet_surface__elevation', grid_units=('km', 'km'),
                symmetric_cbar=True, cmap='Paired')


if __name__ == "__main__":
    main()



#! /usr/bin/env python
"""
Unit tests for landlab.collections
"""

import unittest

from landlab import Palette, Arena, NoProvidersError
from landlab.components.sample import Sample1, Sample2

class TestLandlabPalette(unittest.TestCase):
    def test_empty_palette(self):
        """
        Create a palette without components.
        """
        palette = Palette()

        self.assertEqual(len(palette), 0)
        self.assertEqual(palette.list(), [])
        self.assertEqual(palette.keys(), [])
        self.assertEqual(palette.uses(), [])
        self.assertEqual(palette.provides(), [])

        providers = palette.find_provider('air__temperature')
        self.assertEqual(providers, [])

        users = palette.find_user('air__temperature')
        self.assertEqual(users, [])

        connections = palette.find_connections()
        self.assertDictEqual(connections, {})


class TestLandlabOneComponentPalette(unittest.TestCase):
    def test_create(self):
        palette = Palette(sample=Sample1)

        self.assertEqual(len(palette), 1)

    def test_dict_interface(self):
        palette = Palette(sample=Sample1)

        self.assertDictEqual(dict(sample=Sample1), palette)
        self.assertEqual(len(palette), 1)
        self.assertEqual(palette.keys(), ['sample'])
        self.assertEqual(palette.values(), [Sample1])

        items = palette.items()
        self.assertTupleEqual(('sample', Sample1), items[0])

    def test_list(self):
        palette = Palette(sample=Sample1)

        self.assertEqual(['sample'], palette.list())

    def test_uses(self):
        palette = Palette(sample=Sample1)

        uses = palette.uses()
        uses.sort()
        self.assertEqual(['air__temperature', 'surface__elevation'], uses)

    def test_provides(self):
        palette = Palette(sample=Sample1)

        provides = palette.provides()
        provides.sort()
        self.assertEqual(['deposition__rate'], provides)

    def test_find_providers(self):
        palette = Palette(sample=Sample1)

        providers = palette.find_provider('air__temperature')
        self.assertEqual(providers, [])

        providers = palette.find_provider('deposition__rate')
        self.assertEqual(providers, ['sample'])

    def test_find_users(self):
        palette = Palette(sample=Sample1)

        users = palette.find_user('air__temperature')
        self.assertEqual(users, ['sample'])

    def test_find_connections(self):
        palette = Palette(sample=Sample1)

        with self.assertRaises(NoProvidersError):
            connections = palette.find_connections()


class TestLandlabTwoComponentPalette(unittest.TestCase):
    def test_create(self):
        palette = Palette(one=Sample1, two=Sample2)

        self.assertEqual(len(palette), 2)

    def test_dict_interface(self):
        palette = Palette(one=Sample1, two=Sample2)

        self.assertDictEqual(dict(one=Sample1, two=Sample2), palette)
        self.assertEqual(len(palette), 2)

        keys = palette.keys()
        keys.sort()
        self.assertListEqual(['one', 'two'], keys)

        values = palette.values()
        self.assertEqual(2, len(values))
        self.assertTrue(Sample1 in values and Sample2 in values)

        items = palette.items()
        items.sort()
        self.assertEqual(2, len(items))
        self.assertTupleEqual(('one', Sample1), items[0])
        self.assertTupleEqual(('two', Sample2), items[1])

    def test_list(self):
        palette = Palette(one=Sample1, two=Sample2)

        list = palette.list()
        list.sort()
        self.assertListEqual(['one', 'two'], list)

    def test_uses(self):
        palette = Palette(one=Sample1, two=Sample2)

        uses = palette.uses()
        uses.sort()
        self.assertListEqual(['air__temperature',
                              'deposition__rate',
                              'surface__elevation'], uses)

    def test_provides(self):
        palette = Palette(one=Sample1, two=Sample2)

        provides = palette.provides()
        provides.sort()
        self.assertListEqual(['air__temperature',
                              'deposition__rate',
                              'surface__elevation'], provides)

    def test_find_providers(self):
        palette = Palette(one=Sample1, two=Sample2)

        providers = palette.find_provider('air__temperature')
        self.assertListEqual(['two'], providers)

        providers = palette.find_provider('deposition__rate')
        self.assertEqual(['one'], providers)

    def test_find_users(self):
        palette = Palette(one=Sample1, two=Sample2)

        users = palette.find_user('air__temperature')
        self.assertListEqual(['one'], users)

    def test_find_connections(self):
        palette = Palette(one=Sample1, two=Sample2)

        connections = {
            'one': {'deposition__rate': ['two']},
            'two': {'air__temperature': ['one'],
                    'surface__elevation': ['one']},
        }
        self.assertDictEqual(connections, palette.find_connections())


class TestLandlabArena(unittest.TestCase):
    def test_instantiate(self):
        arena = Arena()
        self.assertDictEqual(dict(), arena)

        arena.instantiate(Sample1, 'one')
        self.assertEqual(1, len(arena))
        self.assertTrue('one' in arena)
        self.assertTrue(isinstance(arena['one'], Sample1))

        arena.instantiate(Sample2, 'two')
        self.assertEqual(2, len(arena))
        self.assertTrue('one' in arena and 'two' in arena)
        self.assertTrue(isinstance(arena['one'], Sample1))
        self.assertTrue(isinstance(arena['two'], Sample2))

    def test_connect(self):
        arena = Arena()
        arena.instantiate(Sample1, 'one')
        arena.instantiate(Sample2, 'two')

        arena.connect('one', 'two', 'air__temperature')
        arena.connect('one', 'two', 'surface__elevation')
        arena.connect('two', 'one', 'deposition__rate')

        dz = arena['one'].get_value('deposition__rate')
        t = arena['two'].get_value('air__temperature')
        z = arena['two'].get_value('surface__elevation')

    def test_walk(self):
        arena = Arena()
        arena.instantiate(Sample1, 'one')
        arena.instantiate(Sample2, 'two')

        arena.connect('one', 'two', 'air__temperature')
        arena.connect('one', 'two', 'surface__elevation')
        arena.connect('two', 'one', 'deposition__rate')

        tree = arena.walk('one')
        self.assertListEqual(['one', 'two'], tree)

if __name__ == '__main__':
    unittest.main()



#! /usr/env/python
"""

Component that models 2D diffusion using an explicit finite-volume method.

Created July 2013 GT
Last updated August 2013 GT

"""

from landlab import ModelParameterDictionary
from landlab import create_and_initialize_grid

_ALPHA = 0.1   # time-step stability factor

_VERSION = 'make_all_data'
#_VERSION = 'explicit'

class DiffusionComponent():
    
    def __init__(self, grid=None, current_time=0.):
        
        self.grid = grid
        self.current_time = current_time
        
    def initialize(self, input_stream):
        
        # Create a ModelParameterDictionary for the inputs
        if type(input_stream)==ModelParameterDictionary:
            inputs = input_stream
        else:
            inputs = ModelParameterDictionary(input_stream)
        
        # Read input/configuration parameters
        self.kd = inputs.get('DIFMOD_KD', ptype=float)
        self.uplift_rate = inputs.get('DIFMOD_UPLIFT_RATE', 0., ptype=float)
        
        # Create grid if one doesn't already exist
        if self.grid==None:
            self.grid = create_and_initialize_grid(input_stream)
            
        # Set internal time step
        # ..todo:
        #   implement mechanism to compute time-steps dynamically if grid is 
        #   adaptive/changing
        dx = self.grid.min_active_link_length()  # smallest active link length
        self.dt = _ALPHA*dx*dx/self.kd  # CFL condition
        
        # Get a list of interior cells
        self.interior_cells = self.grid.get_active_cell_node_ids()
        
        # Here we're experimenting with different approaches: with 
        # 'make_all_data', we create and manage all the data we need and embed
        # it all in the grid. With 'explicit', we require the caller/user to 
        # provide data.
        if _VERSION=='make_all_data':
            #print 'creating internal data'
            self.z = self.grid.add_zeros('node', 'landscape_surface__elevation')
            self.g = self.grid.add_zeros('active_link', 'landscape_surface__gradient')  # surface gradients
            self.qs = self.grid.add_zeros('active_link','unit_sediment_flux')  # unit sediment flux
            self.dqds = self.grid.add_zeros('node', 'sediment_flux_divergence')  # sed flux derivative
        elif _VERSION=='explicit':
            pass
        else:
            # Create data arrays for variables that won't (?) be shared with other
            # components
            self.g = self.grid.create_active_link_array_zeros()  # surface gradients
            self.qs = self.grid.create_active_link_array_zeros()  # unit sediment flux
            self.dqds = self.grid.create_node_array_zeros()  # sed flux derivative
        
        
    def run_one_step_explicit(self, mg, z, g, qs, dqsds, dzdt, delt):
        
        # Take the smaller of delt or built-in time-step size self.dt
        dt = min(self.dt, delt)
        
        # Calculate the gradients and sediment fluxes
        g = mg.calculate_gradients_at_active_links(z)
        qs = -self.kd*g
        
        # Calculate the net deposition/erosion rate at each node
        dqsds = mg.calculate_flux_divergence_at_nodes(qs)
        
        # Calculate the total rate of elevation change
        dzdt = self.uplift_rate - dqsds
        
        # Update the elevations
        z[self.interior_cells] = z[self.interior_cells] \
                                 + dzdt[self.interior_cells] * dt
                                 
        # Update current time and return it
        self.current_time += dt
        
        return z, g, qs, dqsds, dzdt
        
        
    def run_one_step_internal(self, delt):
        
        # Take the smaller of delt or built-in time-step size self.dt
        dt = min(self.dt, delt)
        
        # Calculate the gradients and sediment fluxes
        self.g = self.grid.calculate_gradients_at_active_links(self.z)
        self.qs = -self.kd*self.g
        
        # Calculate the net deposition/erosion rate at each node
        self.dqsds = self.grid.calculate_flux_divergence_at_nodes(self.qs)
        
        # Calculate the total rate of elevation change
        dzdt = self.uplift_rate - self.dqsds
        
        # Update the elevations
        self.z[self.interior_cells] += dzdt[self.interior_cells] * dt
                                 
        # Update current time and return it
        self.current_time += dt
        return self.current_time
        
        
    def run_until_explicit(self, mg, t, z, g, qs, dqsds, dzdt):
        
        while self.current_time < t:
            remaining_time = t - self.current_time
            z, g, qs, dqsds, dzdt = self.run_one_step_explicit(mg, z, g, qs, dqsds, dzdt, remaining_time)
            
        return z, g, qs, dqsds, dzdt
        
        
    def run_until_internal(self, t):
        
        while self.current_time < t:
            remaining_time = t - self.current_time
            self.run_one_step_internal(remaining_time)
                    
        
    def run_until(self, t):  # this is just a temporary duplicate
        
        while self.current_time < t:
            remaining_time = t - self.current_time
            self.run_one_step_internal(remaining_time)
                    
        
    def get_time_step(self):
        """
        Returns time-step size.
        """
        return self.dt
        
